# Copyright 2022 Allen Synthesis
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from europi import *
import machine
from time import ticks_diff, ticks_ms
from random import randint, uniform
from europi_script import EuroPiScript
import gc

"""
Consequencer
author: Nik Ansell (github.com/gamecat69)
date: 2022-02-05
labels: sequencer, triggers, drums, randomness

Consequencer is a gate and stepped CV sequencer inspired by Grids from Mutable Instruments and the randomness created by the Turing Machine.
Pattern changes and randomness are introduced as a consequence of both manual input and control voltages sent to the analogue input.
A large number of popular gate patterns are pre-loaded. Stepped CV sequences are automatically generated.

"""

"""
Version History
March-23    decreased maxRandomPatterns to 32 to avoid crashes on some systems
            pattern is now sum of ain and k2
            randomness is now sum of ain and k1
            added garbage collection to avoid memory allocation errors when creating new random sequences
            scroll pattern on display
            minor pattern updates and reshuffled the order

Jan-24      reduced number of calls to update screen to improve performance for incoming clocks < 7ms
            added grids patterns as an easter egg
            added methods to reduce hysteresis on ain - this reduces the number of ain reads too
            added methods to reduce hysteresis on k1 and k2 - this reduces the number of knob reads too
            added screen saver to improve performance
            changed the way the mode is displayed - replaced M1,M2,M3 with Mr,Mp,Mc for easier reading
            cleaned up some code comments
            added constants for easier code reading
"""

# Operating modes for the internal state machine
MODE_RANDOM = 1
MODE_PATTERN = 2
MODE_CV_PATTERN = 3

# Change detection thresholds to trigger a value change
KNOB_CHANGE_TOLERANCE = 0.999
AIN_CHANGE_TOLERANCE = 1

# AIN Input Mode display chars
MODE_DISPLAY_CHARS = ["r", "p", "c"]

# Wake the screen upon detecting input at ain?
WAKE_SCREEN_ON_AIN_INPUT = False


class Consequencer(EuroPiScript):
    def __init__(self):

        self.loadState()

        self.initPatterns()

        # Initialize variables
        self.step = 0
        self.trigger_duration_ms = 50
        self.clock_step = 0
        self.pattern = 0
        self.pattern_prev = 0
        self.minAnalogInputVoltage = 0.5
        self.randomness = 0
        self.randomness_prev = 0
        self.CvPattern = 0
        self.CvPattern_prev = 0
        self.reset_timeout = 1000
        self.maxRandomPatterns = 32  # This prevents a memory allocation error
        self.maxCvVoltage = clamp(europi_config.MAX_OUTPUT_VOLTAGE, 0, 9)  # The maximum is 9 to maintain single digits in the voltage list
        self.gateVoltage = europi_config.GATE_VOLTAGE
        self.gateVoltages = [0, self.gateVoltage]

        self.ainVal = 0
        self.ainValTemp = 0
        self.k1Val = 0
        self.k1ValTemp = 0
        self.k2Val = 0
        self.k2ValTemp = 0

        self.lastInteractionTimeMs = ticks_ms()
        self.screenOff = False
        self.screenOffTimeoutMs = 10 * 1000

        # State flag to determine if UI state has changed and display should update.
        self._updateUI = True

        # Calculate the longest pattern length to be used when generating random sequences
        self.maxStepLength = len(max(self.BD, key=len))

        # Generate random CV for cv4-6
        self.random4 = []
        self.random5 = []
        self.random6 = []
        self.generateNewRandomCVPattern()

        # Triggered when button 2 is released.
        @b2.handler_falling
        def b2Pressed():
            # grids patterns easter egg
            if (
                ticks_diff(ticks_ms(), b2.last_pressed()) > 3000
                and ticks_diff(ticks_ms(), b2.last_pressed()) < 5000
            ):
                self.gridsMode = not self.gridsMode
                self.saveState()
                self.initPatterns()

            elif ticks_diff(ticks_ms(), b2.last_pressed()) > 300:
                if self.analogInputMode < MODE_CV_PATTERN:
                    self.analogInputMode += 1
                else:
                    self.analogInputMode = MODE_RANDOM
                self.saveState()
            else:
                if self.analogInputMode == MODE_CV_PATTERN:  # Allow changed by CV only in mode 3
                    return

                if self.CvPattern < len(self.random4) - 1:  # change to next CV pattern
                    self.CvPattern += 1
                else:
                    if (
                        len(self.random4) < self.maxRandomPatterns
                    ):  # We need to try and generate a new CV value
                        if self.generateNewRandomCVPattern():
                            self.CvPattern += 1
            self._updateUI = True
            self.screenOff = False
            self.lastInteractionTimeMs = ticks_ms()

        # Triggered when button 1 is released
        @b1.handler_falling
        def b1Pressed():
            if (
                ticks_diff(ticks_ms(), b1.last_pressed()) > 3000
                and ticks_diff(ticks_ms(), b1.last_pressed()) < 5000
            ):
                self.output4isClock = not self.output4isClock
                self.saveState()
            elif ticks_diff(ticks_ms(), b1.last_pressed()) > 300:
                self.random_HH = not self.random_HH
                self.saveState()
            else:
                # Play previous CV Pattern, unless we are at the first pattern
                if self.CvPattern != 0:
                    self.CvPattern -= 1
            self._updateUI = True
            self.screenOff = False
            self.lastInteractionTimeMs = ticks_ms()

        # Triggered on each clock into digital input. Output triggers.
        @din.handler
        def clockTrigger():

            # function timing code. Leave in and activate as needed
            # t = time.ticks_us()

            self.step_length = len(self.BD[self.pattern])

            # A pattern was selected which is shorter than the current step. Set to zero to avoid an error
            if self.step >= self.step_length:
                self.step = 0
            cv5.voltage(self.random5[self.CvPattern][self.step])
            cv6.voltage(self.random6[self.CvPattern][self.step])

            # How much randomness to add to cv1-3
            # As the randomness value gets higher, the chance of a randomly selected int being lower gets higher
            # The output will only trigger if the randint() is <= than the probability of the step in BdProb, SnProb and HhProb respectively
            # Random number 0-99
            randomNumber0_99 = randint(0, 99)
            # Random number 0-9
            randomNumber0_9 = randomNumber0_99 // 10
            if randomNumber0_99 < self.randomness:
                if randomNumber0_9 <= int(self.BdProb[self.pattern][self.step]):
                    cv1.voltage(self.gateVoltages[randint(0, 1)])
                if randomNumber0_9 <= int(self.SnProb[self.pattern][self.step]):
                    cv2.voltage(self.gateVoltages[randint(0, 1)])
                if randomNumber0_9 <= int(self.HhProb[self.pattern][self.step]):
                    cv3.voltage(self.gateVoltages[randint(0, 1)])
            else:
                if randomNumber0_9 <= int(self.BdProb[self.pattern][self.step]):
                    cv1.voltage(self.gateVoltages[int(self.BD[self.pattern][self.step])])
                if randomNumber0_9 <= int(self.SnProb[self.pattern][self.step]):
                    cv2.voltage(self.gateVoltages[int(self.SN[self.pattern][self.step])])

                # If randomize HH is ON:
                if self.random_HH:
                    cv3.value(randint(0, 1))
                else:
                    if randomNumber0_9 <= int(self.HhProb[self.pattern][self.step]):
                        cv3.voltage(self.gateVoltages[int(self.HH[self.pattern][self.step])])

            # Set cv4-6 voltage outputs based on previously generated random pattern
            if self.output4isClock:
                cv4.voltage(self.gateVoltage)
            else:
                cv4.voltage(self.random4[self.CvPattern][self.step])

            # Incremenent the clock step
            self.clock_step += 1
            self.step += 1

            # Update the UI
            if not self.screenOff:
                self._updateUI = True

            # function timing code. Leave in and activate as needed
            # delta = time.ticks_diff(time.ticks_us(), t)
            # print('Function {} Time = {:6.3f}ms'.format('clockTrigger', delta/1000))

        @din.handler_falling
        def clockTriggerEnd():
            cv1.off()
            cv2.off()
            cv3.off()
            if self.output4isClock:
                cv4.off()

    def initPatterns(self):

        # Initialize sequencer pattern arrays
        p = pattern()

        if self.gridsMode:

            self.BD = p.BDGrids
            self.SN = p.SNGrids
            self.HH = p.HHGrids

            # Initialize sequencer pattern probabiltiies
            self.BdProb = p.BdProbGrids
            self.SnProb = p.SnProbGrids
            self.HhProb = p.HhProbGrids

        else:

            self.BD = p.BD
            self.SN = p.SN
            self.HH = p.HH

            # Initialize sequencer pattern probabiltiies
            self.BdProb = p.BdProb
            self.SnProb = p.SnProb
            self.HhProb = p.HhProb

        # Load and populate probability patterns
        # If the probability string len is < pattern len, automatically fill out with the last digit:
        # - 9   becomes 999999999
        # - 95  becomes 955555555
        # - 952 becomes 952222222
        for pi in range(len(self.BD)):
            if len(self.BdProb[pi]) < len(self.BD[pi]):
                self.BdProb[pi] = self.BdProb[pi] + (
                    self.BdProb[pi][-1] * (len(self.BD[pi]) - len(self.BdProb[pi]))
                )
        for pi in range(len(self.SN)):
            if len(self.SnProb[pi]) < len(self.SN[pi]):
                self.SnProb[pi] = self.SnProb[pi] + (
                    self.SnProb[pi][-1] * (len(self.SN[pi]) - len(self.SnProb[pi]))
                )
        for pi in range(len(self.HH)):
            if len(self.HhProb[pi]) < len(self.HH[pi]):
                self.HhProb[pi] = self.HhProb[pi] + (
                    self.HhProb[pi][-1] * (len(self.HH[pi]) - len(self.HhProb[pi]))
                )

    """ Save working vars to a save state file"""

    def saveState(self):
        self.state = {
            "analogInputMode": self.analogInputMode,
            "random_HH": self.random_HH,
            "output4isClock": self.output4isClock,
            "gridsMode": self.gridsMode,
        }
        self.save_state_json(self.state)

    """ Load a previously saved state, or initialize working vars, then save"""

    def loadState(self):
        self.state = self.load_state_json()
        self.analogInputMode = self.state.get("analogInputMode", 1)
        self.random_HH = self.state.get("random_HH", False)
        self.output4isClock = self.state.get("output4isClock", False)
        self.gridsMode = self.state.get("gridsMode", False)
        self.saveState()

    def generateNewRandomCVPattern(self):
        try:
            gc.collect()
            self.random4.append(
                self.generateRandomPattern(self.maxStepLength, 0, self.maxCvVoltage)
            )
            self.random5.append(
                self.generateRandomPattern(self.maxStepLength, 0, self.maxCvVoltage)
            )
            self.random6.append(
                self.generateRandomPattern(self.maxStepLength, 0, self.maxCvVoltage)
            )
            return True
        except Exception:
            return False

    def getPattern(self):
        # If mode 2 and there is CV on the analogue input use it, if not use the knob position
        if self.analogInputMode == MODE_PATTERN and self.ainVal > self.minAnalogInputVoltage:
            self.pattern = min(
                int((len(self.BD) / 100) * self.ainVal) + self.k2Val, len(self.BD) - 1
            )
        else:
            self.pattern = self.k2Val

        self.step_length = len(self.BD[self.pattern])

        if self.pattern_prev != self.pattern:
            self.pattern_prev = self.pattern
            if not self.screenOff:
                self._updateUI = True

    def getCvPattern(self):
        # If analogue input mode 3, get the CV pattern from CV input
        if self.analogInputMode == MODE_CV_PATTERN and self.ainVal > self.minAnalogInputVoltage:
            # Convert percentage value to a representative index of the pattern array
            self.CvPattern = int((len(self.random4) / 100) * self.ainVal)

        if self.CvPattern_prev != self.CvPattern:
            self.CvPattern_prev = self.CvPattern

            if not self.screenOff:
                self._updateUI = True

    def generateRandomPattern(self, length, min, max):
        self.t = []
        for i in range(0, length):
            self.t.append(uniform(0, 9))
        return self.t

    def getRandomness(self):
        # If mode 1 and there is CV on the analogue input use it, if not use the knob position
        if self.analogInputMode == MODE_RANDOM and self.ainVal > self.minAnalogInputVoltage:
            self.randomness = min(self.ainVal + self.k1Val, 99)
        else:
            self.randomness = self.k1Val

        if self.randomness_prev != self.randomness:
            self.randomness_prev = self.randomness

            if not self.screenOff:
                self._updateUI = True

    def getAinVal(self):
        # Read ain val and update if > threshold
        self.ainValTemp = 100 * ain.percent()
        if abs(self.ainValTemp - self.ainVal) > AIN_CHANGE_TOLERANCE:
            self.ainVal = self.ainValTemp


    def getKnobVals(self):
        # Read knob vals and update if > threshold
        self.k1ValTemp = k1.read_position()
        if abs(self.k1ValTemp - self.k1Val) > KNOB_CHANGE_TOLERANCE:
            self.k1Val = self.k1ValTemp
            self.screenOff = False
            self.lastInteractionTimeMs = ticks_ms()

        self.k2ValTemp = k2.read_position(len(self.BD))
        if abs(self.k2ValTemp - self.k2Val) > KNOB_CHANGE_TOLERANCE:
            self.k2Val = self.k2ValTemp
            self.screenOff = False
            self.lastInteractionTimeMs = ticks_ms()

    def main(self):
        while True:
            self.getAinVal()
            self.getKnobVals()
            self.getPattern()
            self.getRandomness()
            self.getCvPattern()
            # Update screen if updateUI flag has been set
            if self._updateUI:
                self.updateScreen()
                self._updateUI = False
            # If I have been running, then stopped for longer than reset_timeout, reset the steps and clock_step to 0
            if (
                self.clock_step != 0
                and ticks_diff(ticks_ms(), din.last_triggered()) > self.reset_timeout
            ):
                self.step = 0
                self.clock_step = 0

            # Has the module been left along for a while, turn off the screen
            if (
                not self.screenOff
                and ticks_diff(ticks_ms(), self.lastInteractionTimeMs) > self.screenOffTimeoutMs
            ):
                self.drawBlankScreen()
                self.screenOff = True

    def visualizePattern(self, pattern, prob):
        output = ""
        for s in range(len(pattern)):
            if pattern[s] == "1":
                char = "^" if prob[s] == "9" else "-"
                output = output + char
            else:
                output = output + " "
        return output

    def drawBlankScreen(self):
        oled.fill(0)
        oled.show()

    def updateScreen(self):
        # oled.clear() - dont use this, it causes the screen to flicker!
        oled.fill(0)

        # Show selected pattern visually

        # Calculate the length of the current pattern
        current_pattern_length = len(self.BD[self.pattern])

        # Calculate the width of one full pattern in pixels
        lpos_offset = current_pattern_length * CHAR_WIDTH

        # Calculate the x position of the first pattern to be drawn
        normal_lpos = lpos = 8 - (self.step * 8)

        # Calculate the number of patterns required to fill the OLED width
        number_of_offset_patterns = 2 * max(int(OLED_WIDTH / lpos_offset), 1)

        # Draw as many offset patterns as required to fill the OLED
        for pattern_offset in range(number_of_offset_patterns):
            # Draw the current pattern
            oled.text(
                self.visualizePattern(self.BD[self.pattern], self.BdProb[self.pattern]),
                normal_lpos,
                0,
                1,
            )
            oled.text(
                self.visualizePattern(self.SN[self.pattern], self.SnProb[self.pattern]),
                normal_lpos,
                10,
                1,
            )
            oled.text(
                self.visualizePattern(self.HH[self.pattern], self.HhProb[self.pattern]),
                normal_lpos,
                20,
                1,
            )
            normal_lpos += lpos_offset

        # If the random toggle is on, show a rectangle
        if self.random_HH:
            oled.fill_rect(0, 29, 10, 3, 1)

        # Show self.output4isClock indicator
        if self.output4isClock:
            oled.rect(12, 29, 10, 3, 1)

        # Show randomness
        oled.text("R" + str(int(self.randomness)), 26, 25, 1)

        # Show CV pattern
        oled.text("C" + str(self.CvPattern), 56, 25, 1)

        # Show the analogInputMode
        oled.text("M" + str(MODE_DISPLAY_CHARS[self.analogInputMode - 1]), 85, 25, 1)

        # Show the pattern number
        if self.gridsMode:
            oled.text(".", 102, 25, 1)

        oled.text(str(self.pattern), 110, 25, 1)

        oled.show()


class pattern:

    # Initialize pattern lists
    BD = []
    SN = []
    HH = []

    BDGrids = []
    SNGrids = []
    HHGrids = []

    # Initialize pattern probabilities

    BdProb = []
    SnProb = []
    HhProb = []

    BdProbGrids = []
    SnProbGrids = []
    HhProbGrids = []

    # 11 interesting patterns
    BD.append("1000100010001000")
    SN.append("0000000000000000")
    HH.append("0000000000000000")
    BdProb.append("9")
    SnProb.append("9")
    HhProb.append("9")

    BD.append("1000100010001000")
    SN.append("0000000000000000")
    HH.append("0010010010010010")
    BdProb.append("9")
    SnProb.append("9")
    HhProb.append("9")

    BD.append("1000100010001000")
    SN.append("0000100000000000")
    HH.append("0010010010010010")
    BdProb.append("9")
    SnProb.append("9")
    HhProb.append("9")

    BD.append("1000100010001000")
    SN.append("0000100000001000")
    HH.append("0010010010010010")
    BdProb.append("9")
    SnProb.append("9")
    HhProb.append("9")

    BD.append("1000100010001000")
    SN.append("0000100000000000")
    HH.append("0000000000000000")
    BdProb.append("9")
    SnProb.append("9")
    HhProb.append("9")

    BD.append("1000100010001000")
    SN.append("0000100000001000")
    HH.append("0000000000000000")
    BdProb.append("9")
    SnProb.append("9")
    HhProb.append("9")

    BD.append("1000100010001000")
    SN.append("0000100000001000")
    HH.append("0000100010001001")
    BdProb.append("9")
    SnProb.append("9")
    HhProb.append("9")

    BD.append("1000100010001000")
    SN.append("0000100000001000")
    HH.append("1010101010101010")
    BdProb.append("9")
    SnProb.append("9")
    HhProb.append("9")

    BD.append("1000100010001000")
    SN.append("0000000000000000")
    HH.append("1111111111111111")
    BdProb.append("9")
    SnProb.append("9")
    HhProb.append("9")

    BD.append("1000100010001000")
    SN.append("0000100000001000")
    HH.append("1111111111111111")
    BdProb.append("9")
    SnProb.append("9")
    HhProb.append("9")

    BD.append("1000100010001000")
    SN.append("0000100000000000")
    HH.append("0001001000000000")
    BdProb.append("9")
    SnProb.append("9")
    HhProb.append("9")

    # 10 commonly found patterns
    # Source: https://docs.google.com/spreadsheets/d/19_3BxUMy3uy1Gb0V8Wc-TcG7q16Amfn6e8QVw4-HuD0/edit#gid=0
    BD.append("1000000010000000")
    SN.append("0000100000001000")
    HH.append("1010101010101010")
    BdProb.append("9")
    SnProb.append("9")
    HhProb.append("9")

    BD.append("1010001000100100")
    SN.append("0000100101011001")
    HH.append("0000000100000100")
    BdProb.append("9")
    SnProb.append("9")
    HhProb.append("9")

    BD.append("1000000110000010")
    SN.append("0000100000001000")
    HH.append("1010101110001010")
    BdProb.append("9")
    SnProb.append("9")
    HhProb.append("9")

    BD.append("1100000100110000")
    SN.append("0000100000001000")
    HH.append("1010101010101010")
    BdProb.append("9")
    SnProb.append("9")
    HhProb.append("9")

    BD.append("1000000110100000")
    SN.append("0000100000001000")
    HH.append("0010101010101010")
    BdProb.append("9")
    SnProb.append("9")
    HhProb.append("9")

    BD.append("1010000000110001")
    SN.append("0000100000001000")
    HH.append("1010101010101010")
    BdProb.append("9")
    SnProb.append("9")
    HhProb.append("9")

    BD.append("1000000110100001")
    SN.append("0000100000001000")
    HH.append("0000100010101011")
    BdProb.append("9")
    SnProb.append("9")
    HhProb.append("9")

    BD.append("1001001010000000")
    SN.append("0000100000001000")
    HH.append("0000100000001000")
    BdProb.append("9")
    SnProb.append("9")
    HhProb.append("9")

    BD.append("1010001001100000")
    SN.append("0000100000001000")
    HH.append("1010101010001010")
    BdProb.append("9")
    SnProb.append("9")
    HhProb.append("9")

    BD.append("1010000101110001")
    SN.append("0000100000001000")
    HH.append("1010101010001010")
    BdProb.append("9")
    SnProb.append("9")
    HhProb.append("9")

    # 5 interesting patterns?
    BD.append("1000100010001000")
    SN.append("0000101001001000")
    HH.append("1010101010101010")
    BdProb.append("9")
    SnProb.append("9")
    HhProb.append("9")

    BD.append("1100000001010000")
    SN.append("0000101000001000")
    HH.append("0101010101010101")
    BdProb.append("9")
    SnProb.append("9")
    HhProb.append("9")

    BD.append("1100000001010000")
    SN.append("0000101000001000")
    HH.append("1111111111111111")
    BdProb.append("9")
    SnProb.append("9")
    HhProb.append("9")

    BD.append("1001001001000100")
    SN.append("0001000000010000")
    HH.append("0101110010011110")
    BdProb.append("9")
    SnProb.append("9")
    HhProb.append("9")

    BD.append("1001001001000100")
    SN.append("0001000000010000")
    HH.append("1111111111111111")
    BdProb.append("9")
    SnProb.append("9")
    HhProb.append("9")

    # 5 Mixed probability patterns
    BD.append("10111111111100001011000000110000")
    SN.append("10001000100010001010000001001000")
    HH.append("10101010101010101010101010101010")
    BdProb.append("99992111129999999999999999969999")
    SnProb.append("95")
    HhProb.append("92939495969792939495969792939492")

    BD.append("10111111111100001011000000110000")
    SN.append("10001000100010001010000001001000")
    HH.append("11111111111111111111111111111111")
    BdProb.append("99992222229999999999999999999999")
    SnProb.append("95")
    HhProb.append("44449999555599996666999922229999")

    BD.append("1000100010001000")
    SN.append("0000101001001000")
    HH.append("0101010101010101")
    BdProb.append("999995")
    SnProb.append("5")
    HhProb.append("99995")

    BD.append("1000110010001100")
    SN.append("0000101001001000")
    HH.append("1111111111111111")
    BdProb.append("9999939999999299")
    SnProb.append("9")
    HhProb.append("9293949592939495")

    BD.append("1000100010001000")
    SN.append("0000101000001000")
    HH.append("1111111111111111")
    BdProb.append("9")
    SnProb.append("9999995999999999")
    HhProb.append("9293949592939495")

    # 5 African Patterns
    BD.append("10110000001100001011000000110000")
    SN.append("10001000100010001010100001001010")
    HH.append("00001011000010110000101100001011")
    BdProb.append("9")
    SnProb.append("9")
    HhProb.append("9")

    BD.append("10101010101010101010101010101010")
    SN.append("00001000000010000000100000001001")
    HH.append("10100010101000101010001010100000")
    BdProb.append("9")
    SnProb.append("9")
    HhProb.append("9")

    BD.append("11000000101000001100000010100000")
    SN.append("00001000000010000000100000001010")
    HH.append("10111001101110011011100110111001")
    BdProb.append("9")
    SnProb.append("9")
    HhProb.append("9")

    BD.append("10001000100010001000100010001010")
    SN.append("00100100101100000010010010110010")
    HH.append("10101010101010101010101010101011")
    BdProb.append("9")
    SnProb.append("9")
    HhProb.append("9")

    BD.append("10010100100101001001010010010100")
    SN.append("00100010001000100010001000100010")
    HH.append("01010101010101010101010101010101")
    BdProb.append("9")
    SnProb.append("9")
    HhProb.append("9")

    # 13 patterns with < 16 steps - can sound disjointed when using CV to select the pattern!

    BD.append("10010000010010")
    SN.append("00010010000010")
    HH.append("11100110111011")
    BdProb.append("9")
    SnProb.append("9")
    HhProb.append("9")

    BD.append("1001000001001")
    SN.append("0001001000001")
    HH.append("1110011011101")
    BdProb.append("9")
    SnProb.append("9")
    HhProb.append("9")

    BD.append("100100000100")
    SN.append("000100100000")
    HH.append("111001101110")
    BdProb.append("9")
    SnProb.append("9")
    HhProb.append("9")

    BD.append("10010000010")
    SN.append("00010010000")
    HH.append("11100110111")
    BdProb.append("9")
    SnProb.append("9")
    HhProb.append("9")

    BD.append("10010000010")
    SN.append("00010010000")
    HH.append("11111010011")
    BdProb.append("9")
    SnProb.append("9")
    HhProb.append("9")

    BD.append("1001000010")
    SN.append("0001000000")
    HH.append("1111101101")
    BdProb.append("9")
    SnProb.append("9")
    HhProb.append("9")

    BD.append("100100010")
    SN.append("000100000")
    HH.append("111110111")
    BdProb.append("9")
    SnProb.append("9")
    HhProb.append("9")

    BD.append("10010010")
    SN.append("00010000")
    HH.append("11111111")
    BdProb.append("9")
    SnProb.append("9")
    HhProb.append("9")

    BD.append("1001001")
    SN.append("0001000")
    HH.append("1111111")
    BdProb.append("9")
    SnProb.append("9")
    HhProb.append("9")

    BD.append("100100")
    SN.append("000100")
    HH.append("111111")
    BdProb.append("9")
    SnProb.append("9")
    HhProb.append("9")

    BD.append("10000")
    SN.append("00001")
    HH.append("11110")
    BdProb.append("9")
    SnProb.append("9")
    HhProb.append("9")

    BD.append("1000")
    SN.append("0000")
    HH.append("1111")
    BdProb.append("9")
    SnProb.append("9")
    HhProb.append("9595")

    BD.append("100")
    SN.append("000")
    HH.append("111")
    BdProb.append("9")
    SnProb.append("9")
    HhProb.append("9")

    # Grids patterns
    # Node: 0
    # Threshold: 180
    BDGrids.append("10000000000010000000100000000000")
    BdProbGrids.append("9")
    # Threshold: 127
    BDGrids.append("10000010000010000000100000000000")
    BdProbGrids.append("9")
    # Threshold: 180
    SNGrids.append("00000000100000000010000010000000")
    SnProbGrids.append("9")
    # Threshold: 127
    SNGrids.append("00000000100000000010000010001000")
    SnProbGrids.append("9")
    # Threshold: 180
    HHGrids.append("00001000000010000000100000001000")
    HhProbGrids.append("9")
    # Threshold: 127
    HHGrids.append("10001000101010001000100010001000")
    HhProbGrids.append("9")

    # Node: 1
    # Threshold: 180
    BDGrids.append("10000000100000001000000000000000")
    BdProbGrids.append("9")
    # Threshold: 127
    BDGrids.append("10000000100000001000000010001000")
    BdProbGrids.append("9")
    # Threshold: 180
    SNGrids.append("00100000100000000000000010000000")
    SnProbGrids.append("9")
    # Threshold: 127
    SNGrids.append("00100000100010000010000010000000")
    SnProbGrids.append("9")
    # Threshold: 180
    HHGrids.append("00001000000010000000100000001000")
    HhProbGrids.append("9")
    # Threshold: 127
    HHGrids.append("10101000101010001000100010001000")
    HhProbGrids.append("9")

    # Node: 2
    # Threshold: 180
    BDGrids.append("10000000000000100000100000000000")
    BdProbGrids.append("9")
    # Threshold: 127
    BDGrids.append("10000000000000100010100000001000")
    BdProbGrids.append("9")
    # Threshold: 180
    SNGrids.append("00000000101000000000000010000000")
    SnProbGrids.append("9")
    # Threshold: 127
    SNGrids.append("00000000101000000000000010000010")
    SnProbGrids.append("9")
    # Threshold: 180
    HHGrids.append("10000000000000001000100000001000")
    HhProbGrids.append("9")
    # Threshold: 127
    HHGrids.append("10001000000000001010101000001000")
    HhProbGrids.append("9")

    # Node: 3
    # Threshold: 180
    BDGrids.append("10100000000000001000100000000000")
    BdProbGrids.append("9")
    # Threshold: 127
    BDGrids.append("10100000001000001000100000100000")
    BdProbGrids.append("9")
    # Threshold: 180
    SNGrids.append("00000000101000000000000010000000")
    SnProbGrids.append("9")
    # Threshold: 127
    SNGrids.append("00001000101000000000000010001000")
    SnProbGrids.append("9")
    # Threshold: 180
    HHGrids.append("10101010000000001000000000001000")
    HhProbGrids.append("9")
    # Threshold: 127
    HHGrids.append("10101010100000001000000010001000")
    HhProbGrids.append("9")

    # Node: 4
    # Threshold: 180
    BDGrids.append("10000000000010001000000000000000")
    BdProbGrids.append("9")
    # Threshold: 127
    BDGrids.append("10000000000010001000100000000000")
    BdProbGrids.append("9")
    # Threshold: 180
    SNGrids.append("00000000100000000000000010101000")
    SnProbGrids.append("9")
    # Threshold: 127
    SNGrids.append("00000000100000000000100010101000")
    SnProbGrids.append("9")
    # Threshold: 180
    HHGrids.append("00001000000010000000100000000000")
    HhProbGrids.append("9")
    # Threshold: 127
    HHGrids.append("00001000100010000000100010001000")
    HhProbGrids.append("9")

    # Node: 5
    # Threshold: 180
    BDGrids.append("10000000000000001010000000000000")
    BdProbGrids.append("9")
    # Threshold: 127
    BDGrids.append("10000000000000001010001000100000")
    BdProbGrids.append("9")
    # Threshold: 180
    SNGrids.append("00000010100000000000000010000010")
    SnProbGrids.append("9")
    # Threshold: 127
    SNGrids.append("10001010100000000000000010000010")
    SnProbGrids.append("9")
    # Threshold: 180
    HHGrids.append("00001000000010000000100000001000")
    HhProbGrids.append("9")
    # Threshold: 127
    HHGrids.append("10001000000010001000100000001000")
    HhProbGrids.append("9")

    # Node: 6
    # Threshold: 180
    BDGrids.append("10001000000000000000000000001000")
    BdProbGrids.append("9")
    # Threshold: 127
    BDGrids.append("10001000000000000000100000001000")
    BdProbGrids.append("9")
    # Threshold: 180
    SNGrids.append("00100000100000000010000010000000")
    SnProbGrids.append("9")
    # Threshold: 127
    SNGrids.append("00100000100000100010000010100000")
    SnProbGrids.append("9")
    # Threshold: 180
    HHGrids.append("10101010100000000000000000000000")
    HhProbGrids.append("9")
    # Threshold: 127
    HHGrids.append("10101010100010001000000010000000")
    HhProbGrids.append("9")

    # Node: 7
    # Threshold: 180
    BDGrids.append("10000000000010001000000000001000")
    BdProbGrids.append("9")
    # Threshold: 127
    BDGrids.append("10000000000010001000100000001000")
    BdProbGrids.append("9")
    # Threshold: 180
    SNGrids.append("00000000100010000000000010001000")
    SnProbGrids.append("9")
    # Threshold: 127
    SNGrids.append("00000000100010000000100010001000")
    SnProbGrids.append("9")
    # Threshold: 180
    HHGrids.append("10001000100000001000100010000000")
    HhProbGrids.append("9")
    # Threshold: 127
    HHGrids.append("10001000100010001000100010000000")
    HhProbGrids.append("9")

    # Node: 8
    # Threshold: 180
    BDGrids.append("10001000000010001000100000000000")
    BdProbGrids.append("9")
    # Threshold: 127
    BDGrids.append("10001000000010001000100000001000")
    BdProbGrids.append("9")
    # Threshold: 180
    SNGrids.append("00000000100000000000100010001000")
    SnProbGrids.append("9")
    # Threshold: 127
    SNGrids.append("10000000100000000000100010001000")
    SnProbGrids.append("9")
    # Threshold: 180
    HHGrids.append("10000000000000000000000010101010")
    HhProbGrids.append("9")
    # Threshold: 127
    HHGrids.append("10000000000000001000000010101010")
    HhProbGrids.append("9")

    # Node: 9
    # Threshold: 180
    BDGrids.append("10000000000010000000100000100000")
    BdProbGrids.append("9")
    # Threshold: 127
    BDGrids.append("10000010000010000000100000101000")
    BdProbGrids.append("9")
    # Threshold: 180
    SNGrids.append("00000000100000100000100010000000")
    SnProbGrids.append("9")
    # Threshold: 127
    SNGrids.append("00000000100000100000100010000010")
    SnProbGrids.append("9")
    # Threshold: 180
    HHGrids.append("00001010000010001000100000001000")
    HhProbGrids.append("9")
    # Threshold: 127
    HHGrids.append("00001010100010001000101010001000")
    HhProbGrids.append("9")

    # Node: 10
    # Threshold: 180
    BDGrids.append("00000000000010000010000000001000")
    BdProbGrids.append("9")
    # Threshold: 127
    BDGrids.append("10000000000010000010000000001000")
    BdProbGrids.append("9")
    # Threshold: 180
    SNGrids.append("00000000001010001000000000000000")
    SnProbGrids.append("9")
    # Threshold: 127
    SNGrids.append("00001000101010001000000000000000")
    SnProbGrids.append("9")
    # Threshold: 180
    HHGrids.append("10000000100000001000000010001000")
    HhProbGrids.append("9")
    # Threshold: 127
    HHGrids.append("10001000100000001000000010001000")
    HhProbGrids.append("9")

    # Node: 11
    # Threshold: 180
    BDGrids.append("10000000000000001000100000100000")
    BdProbGrids.append("9")
    # Threshold: 127
    BDGrids.append("10000000000000001000100000100010")
    BdProbGrids.append("9")
    # Threshold: 180
    SNGrids.append("00000000100100100000000010000000")
    SnProbGrids.append("9")
    # Threshold: 127
    SNGrids.append("00000000100100100000000010001100")
    SnProbGrids.append("9")
    # Threshold: 180
    HHGrids.append("00001000000010001000100010001000")
    HhProbGrids.append("9")
    # Threshold: 127
    HHGrids.append("10001000100010001000100010001000")
    HhProbGrids.append("9")

    # Node: 12
    # Threshold: 180
    BDGrids.append("10001010000000000000101000000000")
    BdProbGrids.append("9")
    # Threshold: 127
    BDGrids.append("10001010000000000000101000001000")
    BdProbGrids.append("9")
    # Threshold: 180
    SNGrids.append("00000000100000100000000010000000")
    SnProbGrids.append("9")
    # Threshold: 127
    SNGrids.append("00000000100000100010000010000010")
    SnProbGrids.append("9")
    # Threshold: 180
    HHGrids.append("10001000100010001000000000000000")
    HhProbGrids.append("9")
    # Threshold: 127
    HHGrids.append("10001000100010001000100010001010")
    HhProbGrids.append("9")

    # Node: 13
    # Threshold: 180
    BDGrids.append("10000000100000101000000000000000")
    BdProbGrids.append("9")
    # Threshold: 127
    BDGrids.append("10000000100000101000000010000000")
    BdProbGrids.append("9")
    # Threshold: 180
    SNGrids.append("00000000100000000000000010100000")
    SnProbGrids.append("9")
    # Threshold: 127
    SNGrids.append("00000000100000000000000010101000")
    SnProbGrids.append("9")
    # Threshold: 180
    HHGrids.append("10001000100010001000100000000000")
    HhProbGrids.append("9")
    # Threshold: 127
    HHGrids.append("10101010100010101000100010001000")
    HhProbGrids.append("9")

    # Node: 14
    # Threshold: 180
    BDGrids.append("10000000000000001000000000000000")
    BdProbGrids.append("9")
    # Threshold: 127
    BDGrids.append("10000000000000001000100000000000")
    BdProbGrids.append("9")
    # Threshold: 180
    SNGrids.append("00000000100010000000000000001010")
    SnProbGrids.append("9")
    # Threshold: 127
    SNGrids.append("00000000100010000000000010001010")
    SnProbGrids.append("9")
    # Threshold: 180
    HHGrids.append("10000000000000001000000000000000")
    HhProbGrids.append("9")
    # Threshold: 127
    HHGrids.append("10001000000000001000100000000000")
    HhProbGrids.append("9")

    # Node: 15
    # Threshold: 180
    BDGrids.append("10000000000010001000000000000000")
    BdProbGrids.append("9")
    # Threshold: 127
    BDGrids.append("10000000000010001000000000001000")
    BdProbGrids.append("9")
    # Threshold: 180
    SNGrids.append("00000000100010000000000010000000")
    SnProbGrids.append("9")
    # Threshold: 127
    SNGrids.append("00000000100010000000000010001000")
    SnProbGrids.append("9")
    # Threshold: 180
    HHGrids.append("10001000100000000000000000000000")
    HhProbGrids.append("9")
    # Threshold: 127
    HHGrids.append("10001000100000001000100000000000")
    HhProbGrids.append("9")

    # Node: 16
    # Threshold: 180
    BDGrids.append("10000000000000001000000010000000")
    BdProbGrids.append("9")
    # Threshold: 127
    BDGrids.append("10000000000000001000000010001000")
    BdProbGrids.append("9")
    # Threshold: 180
    SNGrids.append("00001000000010000000100000000000")
    SnProbGrids.append("9")
    # Threshold: 127
    SNGrids.append("00001000000010000000100000001000")
    SnProbGrids.append("9")
    # Threshold: 180
    HHGrids.append("00000000000000001000100010000000")
    HhProbGrids.append("9")
    # Threshold: 127
    HHGrids.append("10000000000000001000100010001000")
    HhProbGrids.append("9")

    # Node: 17
    # Threshold: 180
    BDGrids.append("10000000100000001000000000000000")
    BdProbGrids.append("9")
    # Threshold: 127
    BDGrids.append("10000000100000001000000010000000")
    BdProbGrids.append("9")
    # Threshold: 180
    SNGrids.append("00000000000000001000001000001000")
    SnProbGrids.append("9")
    # Threshold: 127
    SNGrids.append("00000000100000101000001000001000")
    SnProbGrids.append("9")
    # Threshold: 180
    HHGrids.append("00000000000000000010000010001000")
    HhProbGrids.append("9")
    # Threshold: 127
    HHGrids.append("00001000000010000010000010001000")
    HhProbGrids.append("9")

    # Node: 18
    # Threshold: 180
    BDGrids.append("10000000100000001000000010000000")
    BdProbGrids.append("9")
    # Threshold: 127
    BDGrids.append("10000000100000001000000010101010")
    BdProbGrids.append("9")
    # Threshold: 180
    SNGrids.append("00000000100000000000000000001000")
    SnProbGrids.append("9")
    # Threshold: 127
    SNGrids.append("00000000100000000000000000001000")
    SnProbGrids.append("9")
    # Threshold: 180
    HHGrids.append("10001000000000100010000000000000")
    HhProbGrids.append("9")
    # Threshold: 127
    HHGrids.append("10001000000000100010001000000000")
    HhProbGrids.append("9")

    # Node: 19
    # Threshold: 180
    BDGrids.append("10000010100000000000000000000000")
    BdProbGrids.append("9")
    # Threshold: 127
    BDGrids.append("10000010100000101000000000000000")
    BdProbGrids.append("9")
    # Threshold: 180
    SNGrids.append("00000000000010100000000000001010")
    SnProbGrids.append("9")
    # Threshold: 127
    SNGrids.append("00000000000010100000100000001010")
    SnProbGrids.append("9")
    # Threshold: 180
    HHGrids.append("10000010000000000000000010000000")
    HhProbGrids.append("9")
    # Threshold: 127
    HHGrids.append("10000010000010000000100010000000")
    HhProbGrids.append("9")

    # Node: 20
    # Threshold: 180
    BDGrids.append("10000000100000001000100000000000")
    BdProbGrids.append("9")
    # Threshold: 127
    BDGrids.append("10000000100000001000100010001000")
    BdProbGrids.append("9")
    # Threshold: 180
    SNGrids.append("00000000100000100010000010000000")
    SnProbGrids.append("9")
    # Threshold: 127
    SNGrids.append("00000000101000100010000010000000")
    SnProbGrids.append("9")
    # Threshold: 180
    HHGrids.append("00000000000010000000100000001000")
    HhProbGrids.append("9")
    # Threshold: 127
    HHGrids.append("00001000000010000000100000001000")
    HhProbGrids.append("9")

    # Node: 21
    # Threshold: 180
    BDGrids.append("10001000000000001000000010000000")
    BdProbGrids.append("9")
    # Threshold: 127
    BDGrids.append("10001000100000001000000010000000")
    BdProbGrids.append("9")
    # Threshold: 180
    SNGrids.append("00000000100010000000000000000000")
    SnProbGrids.append("9")
    # Threshold: 127
    SNGrids.append("00000000100010000000000000100000")
    SnProbGrids.append("9")
    # Threshold: 180
    HHGrids.append("00001000000010000010100000001000")
    HhProbGrids.append("9")
    # Threshold: 127
    HHGrids.append("00001000000010001010100000001000")
    HhProbGrids.append("9")

    # Node: 22
    # Threshold: 180
    BDGrids.append("10000000100000100000000010000000")
    BdProbGrids.append("9")
    # Threshold: 127
    BDGrids.append("10000000100000101000000010000000")
    BdProbGrids.append("9")
    # Threshold: 180
    SNGrids.append("10000000100000100000000000000000")
    SnProbGrids.append("9")
    # Threshold: 127
    SNGrids.append("10000000100000100010000010000000")
    SnProbGrids.append("9")
    # Threshold: 180
    HHGrids.append("00001000000010000000000000001000")
    HhProbGrids.append("9")
    # Threshold: 127
    HHGrids.append("00001000000010000010001000001000")
    HhProbGrids.append("9")

    # Node: 23
    # Threshold: 180
    BDGrids.append("10001000101000000000000000000000")
    BdProbGrids.append("9")
    # Threshold: 127
    BDGrids.append("10001000101000001010001010000000")
    BdProbGrids.append("9")
    # Threshold: 180
    SNGrids.append("00000000000010000000001000001000")
    SnProbGrids.append("9")
    # Threshold: 127
    SNGrids.append("00000000000010000000001100001000")
    SnProbGrids.append("9")
    # Threshold: 180
    HHGrids.append("10101111000000001010000000000000")
    HhProbGrids.append("9")
    # Threshold: 127
    HHGrids.append("10101111000000001010111110000000")
    HhProbGrids.append("9")

    # Node: 24
    # Threshold: 180
    BDGrids.append("00000100100000000000010000000000")
    BdProbGrids.append("9")
    # Threshold: 127
    BDGrids.append("10000100100000001000010000000000")
    BdProbGrids.append("9")
    # Threshold: 180
    SNGrids.append("10000000000001000000000010000000")
    SnProbGrids.append("9")
    # Threshold: 127
    SNGrids.append("10000000000001001000010010000100")
    SnProbGrids.append("9")
    # Threshold: 180
    HHGrids.append("10000100000001000000000000000000")
    HhProbGrids.append("9")
    # Threshold: 127
    HHGrids.append("10000100100001000000000000000000")
    HhProbGrids.append("9")


if __name__ == "__main__":
    # Reset module display state.
    turn_off_all_cvs()
    dm = Consequencer()
    dm.main()



# Copyright 2024 Allen Synthesis
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
try:
    # Local development
    from software.firmware import europi_config
    from software.firmware.europi import CHAR_HEIGHT, CHAR_WIDTH, OLED_WIDTH
    from software.firmware.europi import ain, din, k1, k2, oled, b1, b2, cvs
    from software.firmware.europi_script import EuroPiScript
    from software.firmware.experimental.a_to_d import AnalogReaderDigitalWrapper
    from software.firmware.experimental.custom_font import CustomFontDisplay
    from software.firmware.experimental.fonts import ubuntumono20

except ImportError:
    # Device import path
    from europi import *
    from europi_script import EuroPiScript
    from experimental.a_to_d import AnalogReaderDigitalWrapper
    from experimental.custom_font import CustomFontDisplay
    from experimental.fonts import ubuntumono20

from random import random, seed, randint
from time import sleep_ms, ticks_diff, ticks_ms

DEFAULT_SEQUENCE_LENGTH = 16
DEFAULT_SEED = 0x8F26
DEFAULT_PROBABILITIES = [0.92, 0.86, 0.64, 0.48, 0.32, 0.18]
MAX_STEPS = 32
LONG_PRESS_MS = 500

# Use the Custom Font wrapper for oled display.
oled = CustomFontDisplay()


class TriggerMode:
    TRIGGER = 1
    GATE = 2
    FLIP = 3


class Page:
    MAIN = 1
    EDIT_PROBABILITY = 2
    EDIT_SEED = 3


class Output:
    def __init__(self, cv, probability):
        self.cv = cv
        self.probability = probability
        self.pattern = [0] * MAX_STEPS
        self.build_pattern()

    def build_pattern(self):
        for i in range(MAX_STEPS):
            self.pattern[i] = random() < self.probability

    def trigger(self, step, mode):
        """Input trigger has gone high. Change the cv state based on mode for given step."""
        # If random is greater than the probability range, return without triggering.
        if not self.pattern[step]:
            return

        if mode == TriggerMode.FLIP:
            self.cv.toggle()
        else:
            self.cv.on()

    def trigger_off(self):
        """Input trigger has gone low. Turn off when in 'trigger' mode."""
        self.cv.off()


class SeedPacket:
    def __init__(self, seed, outputs, probabilities, mode):
        self.seed = seed
        self.mode = mode
        self.outputs = []
        for cv, prob in zip(outputs, probabilities):
            self.outputs.append(Output(cv, prob))

    def new_seed(self):
        """Seed the pseudo-random number generator with a new random seed and generate new patterns."""
        self.seed = randint(0,0xFFFF)
        seed(self.seed)
        for out in self.outputs:
            out.build_pattern()
        return self.seed

    def trigger(self, step):
        """Input trigger has gone high."""
        if self.mode == TriggerMode.GATE:
            for out in self.outputs:
                out.trigger_off()
            # Brief low period between gates.
            sleep_ms(5)

        for out in self.outputs:
            out.trigger(step, self.mode)

    def trigger_off(self):
        """Input trigger has gone low."""
        if self.mode != TriggerMode.TRIGGER:
            return

        for out in self.outputs:
            out.trigger_off()

    @property
    def probabilities(self):
        """Return the list of output probability between 0 and 1."""
        return [out.probability for out in self.outputs]

    def update_probability(self, index, prob):
        """Update the probability and pattern for the given output."""
        self.outputs[index].probability = prob
        self.outputs[index].build_pattern()


class BitGarden(EuroPiScript):
    def __init__(self):
        super().__init__()
        # Default initial state
        self.state = {
            "sequence_length": DEFAULT_SEQUENCE_LENGTH,
            "seed": DEFAULT_SEED,
            "mode": TriggerMode.TRIGGER,
            "probabilities": DEFAULT_PROBABILITIES,
        }
        self.state.update(self.load_state_json())

        # Class state variables
        self.step_counter = 0
        self.sequence_length = self.state["sequence_length"]
        self.packet = SeedPacket(self.state["seed"], cvs, self.state["probabilities"], self.state["mode"])
        self.pattern_index = 0

        # Private class instance variables
        self._page = Page.MAIN
        self._update_display = True
        self._prev_k1 = None
        self._prev_k2 = None
        self._seed_index = 0
        self._prob_index = 0

        # Attach IRQ Handlers
        din.handler(self.digital_rising)
        din.handler_falling(self.digital_falling)
        b1.handler_falling(self.b1_handler)
        b2.handler_falling(self.b2_handler)
        self.din2 = AnalogReaderDigitalWrapper(ain, cb_rising=self.digital2_rising)


    @classmethod
    def display_name(cls):
        return "Bit Garden"

    def save_state(self):
        """Save the current state variables as JSON."""
        self.state.update({
            "sequence_length": self.sequence_length,
            "seed": self.packet.seed,
            "mode": self.packet.mode,
            "probabilities": self.packet.probabilities,
        })
        self.save_state_json(self.state)

    def digital_rising(self):
        """Advance the pseudo-random sequence step counter."""
        self.step_counter = (self.step_counter + 1) % self.sequence_length
        self.packet.trigger(self.step_counter)
        self._update_display = True

    def digital_falling(self):
        """Tell the SeedPacket the digital input has gone low for Trigger mode."""
        self.packet.trigger_off()

    def digital2_rising(self):
        self.packet.new_seed()
        self._update_display = True

    def b1_handler(self):
        self._update_display = True

        # Short press  to toggle modde, long press to switch to edit probability page.
        if self._page == Page.MAIN:
            if ticks_diff(ticks_ms(), b1.last_pressed()) < LONG_PRESS_MS:
                self.toggle_mode()
            else:
                self._prev_k2 = None
                self._page = Page.EDIT_PROBABILITY

        # If on Edit Probability page, b1 press will return home.
        elif self._page == Page.EDIT_PROBABILITY:
            self._prob_index = (self._prob_index + 1) % 6

        # Change seed edit index.
        elif self._page == Page.EDIT_SEED:
            self._seed_index = (self._seed_index + 1) % 4

    def b2_handler(self):
        self._update_display = True

        if self._page == Page.MAIN:
            # Short press vs Long press
            if ticks_diff(ticks_ms(), b2.last_pressed()) < LONG_PRESS_MS:
                self.new_seed()
            else:
                self._temp_seed = self.packet.seed
                self._prev_k2 = None
                self._page = Page.EDIT_SEED

        elif self._page == Page.EDIT_PROBABILITY:
            self._page = Page.MAIN

        elif self._page == Page.EDIT_SEED:
            # If on Edit Seed page, b2 press will return home.
            self.packet.seed = self._temp_seed
            self._page = Page.MAIN

    def toggle_mode(self):
        """Change the current output mode."""
        if self.packet.mode == TriggerMode.TRIGGER:
            self.packet.mode = TriggerMode.GATE
        elif self.packet.mode == TriggerMode.GATE:
            self.packet.mode = TriggerMode.FLIP
        elif self.packet.mode == TriggerMode.FLIP:
            self.packet.mode = TriggerMode.TRIGGER
        self.save_state()

    def new_seed(self):
        """Generate a new seed to produce a new pseudo-random trigger sequence."""
        self.state.update({"seed": self.packet.new_seed()})
        self.save_state()

    def update_sequence_length(self):
        """Check if the knob has changed value and update the sequence length accordingly."""
        read = k1.range(MAX_STEPS-1) + 2  # range of 2..MAX_STEPS
        if read != self._prev_k1:
            self._prev_k1 = read
            self.sequence_length = read
            self.step_counter = min(self.step_counter, self.sequence_length)
            self._update_display = True

    def update_selected_pattern(self):
        """Change which pattern is currently displayed."""
        index = k2.range(len(self.packet.outputs))
        if index == self._prev_k2 or self._prev_k2 is None:
            self._prev_k2 = index
            return
        self._prev_k2 = index
        self.pattern_index = index
        self._update_display = True

    def update_seed_digit(self):
        """Update the seed digit for selected digit index."""
        value = k2.range(16)
        if value == self._prev_k2 or self._prev_k2 is None:
            self._prev_k2 = value
            return
        self._prev_k2 = value

        if self._seed_index == 0:
            position = 0x1000
            clear = 0x0FFF
        elif self._seed_index == 1:
            position = 0x0100
            clear = 0xF0FF
        elif self._seed_index == 2:
            position = 0x0010
            clear = 0xFF0F
        elif self._seed_index == 3:
            position = 0x0001
            clear = 0xFFF0
        self._temp_seed = (self._temp_seed & clear) + (int(position) * value)
        self._update_display = True

    def update_selected_prob(self):
        value = k2.range(100)
        if value == self._prev_k2 or self._prev_k2 is None:
            self._prev_k2 = value
            return
        self._prev_k2 = value
        self.packet.update_probability(self._prob_index, k2.percent())
        self._update_display = True

    def mode_text(self):
        """Return the display text for current Trigger Mode."""
        if self.packet.mode == TriggerMode.TRIGGER:
            return "TRIG"
        elif self.packet.mode == TriggerMode.GATE:
            return "GATE"
        elif self.packet.mode == TriggerMode.FLIP:
            return "FLIP"

    def update_display(self):
        """If UI state has changed, update the display."""
        if not self._update_display:
            return
        self._update_display = False

        oled.fill(0)
        if self._page == Page.MAIN:
            self.display_main(self.packet.outputs[self.pattern_index].pattern)
        elif self._page == Page.EDIT_PROBABILITY:
            self.display_edit_probability()
        elif self._page == Page.EDIT_SEED:
            self.display_edit_seed()
        oled.show()

        return

    def display_main(self, pattern):
        start = 2
        top = 0
        left = start
        boxSize = 8
        trigSize = 2
        tpadding = 3
        lpadding = 3
        wrap = 8

        for i in range(self.sequence_length):
            _top = top
            if self.sequence_length <= 8:
                _top += 8
            elif self.sequence_length <= 16:
                _top += 6
                tpadding = 2
            elif self.sequence_length <= 24:
                _top += 4
                tpadding = 1
            else:
                tpadding = 0

            # Step boxes
            oled.rect(left, _top, boxSize, boxSize, 1)

            # Fill current step
            if i == self.step_counter:
                oled.rect(left, _top, boxSize, boxSize, 1, 1)

            # Trigger indicator inside the step box
            if pattern[i]:
                if (i == self.step_counter):
                    oled.rect(left+2, _top+2, trigSize+2, trigSize+2, 0, 1)
                else:
                    oled.rect(left+3, _top+3, trigSize, trigSize, 1, 1)

            left += boxSize + lpadding

            # Wrap the box draw cursor if we hit wrap count.
            if (i+1) % wrap == 0:
                top += boxSize + tpadding
                left = start

        oled.text(f"CV:{self.pattern_index+1}", OLED_WIDTH-(CHAR_WIDTH*4), 0, 1)
        oled.text(f"{self.mode_text()}", OLED_WIDTH-(CHAR_WIDTH*4), CHAR_HEIGHT, 1)
        oled.text(f"{self.packet.seed:04X}", OLED_WIDTH-(CHAR_WIDTH*4), CHAR_HEIGHT*2, 1)
        oled.text(f"{self.step_counter+1:>2}/{self.sequence_length:>2} ", OLED_WIDTH-(5*CHAR_WIDTH), CHAR_HEIGHT*3, 1)

    def display_edit_seed(self):
        """Display the editable seed and digit indicator."""
        start = 40
        top = 6
        charw = ubuntumono20.max_width()
        charh = ubuntumono20.height()
        oled.text(f"{self._temp_seed:04X}", start, top, font=ubuntumono20)
        oled.hline(start + (self._seed_index * charw), top+charh, charw, 1)

    def display_edit_probability(self):
        """Display each output probability as a vertical filled bar with edit indicator."""
        top = 0
        left = 16
        barWidth = 10
        barHeight = 30
        padding = 8

        for i, prob in enumerate(self.packet.probabilities):
            # Draw output probability bar.
            oled.rect(left, top, barWidth, barHeight, 1)

            # Fill current bar probability.
            probFill = int(float(barHeight) * prob)
            probTop = top + barHeight - probFill
            oled.rect(left, probTop, barWidth, probFill, 1, 1)

            # Show selected output.
            if i == self._prob_index:
                oled.text(">", left - CHAR_WIDTH, probTop, 1)

            left += barWidth + padding

    def main(self):
        """Main loop for detecing knob changes and updating display."""
        while True:
            self.din2.update()
            # Current page determines which input methods is called.
            if self._page == Page.MAIN:
                self.update_sequence_length()
                self.update_selected_pattern()
            elif self._page == Page.EDIT_PROBABILITY:
                self.update_selected_prob()
            elif self._page == Page.EDIT_SEED:
                self.update_seed_digit()
            self.update_display()


if __name__ == "__main__":
    BitGarden().main()



# Copyright 2024 Allen Synthesis
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from europi import *
from random import random
from time import sleep_ms, ticks_ms, ticks_add, ticks_diff
import machine
from europi_script import EuroPiScript

# Internal clock tempo range.
MAX_BPM = 280
MIN_BPM = 20

# Constant values for display.
FRAME_WIDTH = int(OLED_WIDTH / 8)


class CoinToss(EuroPiScript):
    def __init__(self):
        self.gate_mode = True
        self.internal_clock = True
        self._prev_clock = 0
        self._tempo = 0
        self._deadline = 0

        @b1.handler
        def toggle_clock():
            """Toggle between internal clock and external clock from digital in."""
            self.internal_clock = not self.internal_clock

        @b2.handler
        def toggle_gate():
            """Toggle between gate and trigger mode."""
            self.gate_mode = not self.gate_mode
            turn_off_all_cvs()

    def tempo(self):
        """Read the current tempo set by k1 within set range."""
        return round(k1.read_position(MAX_BPM - MIN_BPM) + MIN_BPM)

    def get_next_deadline(self):
        """Get the deadline for next clock tick whole note."""
        # The duration of a quarter note in ms for the current tempo.
        wait_ms = int(((60 / self.tempo()) / 4) * 1000)
        return ticks_add(ticks_ms(), wait_ms)

    def wait(self):
        """Pause script execution waiting for next quarter note in the clock cycle."""
        if self.internal_clock:
            while True:
                if ticks_diff(self._deadline, ticks_ms()) <= 0:
                    self._deadline = self.get_next_deadline()
                    return
        else:  # External clock
            # Loop until digital in goes high (clock pulse received).
            while not self.internal_clock:
                if din.value() != self._prev_clock:
                    # We've detected a new clock value.
                    self._prev_clock = 1 if self._prev_clock == 0 else 0
                    # If the previous value was just set to 1 then we are seeing
                    # a high value for the first time, break wait and return.
                    if self._prev_clock == 1:
                        return

    def toss(self, a, b, draw=True):
        """If random value is below trigger a, otherwise trigger b.

        If draw is true, then display visualization of the coin toss.
        """
        coin = random()
        # Sum the knob2 and analogue input values to determine threshold.
        read_sum = k2.percent() + ain.read_voltage()/12
        self.threshold = clamp(read_sum, 0, 1)
        if self.gate_mode:
            a.value(coin < self.threshold)
            b.value(coin > self.threshold)
        else:
            (a if coin < self.threshold else b).on()

        if not draw:
            return

        # Draw gate/trigger display graphics for coin toss
        h = int(self.threshold * OLED_HEIGHT)
        tick = FRAME_WIDTH if self.gate_mode else 1
        offset = 8  # The amount of negative space before drawing gate.
        if coin < self.threshold:
            oled.fill_rect(offset, 0, tick, h, 1)
        else:
            oled.fill_rect(offset, h, tick, OLED_HEIGHT, 1)


    def main(self):
        # Start the main loop.
        counter = 0
        while True:
            # Scroll and clear new screen area.
            oled.scroll(FRAME_WIDTH, 0)
            oled.fill_rect(0, 0, FRAME_WIDTH, OLED_HEIGHT, 0)

            self.toss(cv1, cv2)
            cv3.on()  # First column clock trigger
            if counter % 4 == 0:
                self.toss(cv4, cv5, False)
                cv6.on()  # Second column clock trigger (1/4x speed)

            sleep_ms(10)
            if self.gate_mode:
                # Only turn off clock triggers.
                cv3.off()
                cv6.off()
            else:
                # Turn of all cvs in trigger mode.
                turn_off_all_cvs()

            # Draw threshold line
            oled.hline(0, int(self.threshold * OLED_HEIGHT), FRAME_WIDTH, 1)
            oled.show()

            counter += 1
            self.wait()

if __name__ == '__main__':
    # Reset module display state.
    coin_toss = CoinToss()
    coin_toss.main()



#!/usr/bin/env python3
# Copyright 2024 Allen Synthesis
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Arpeggiator & ascending/descending scale generator

Outputs ascending/descending quantized CV, with additional outputs
a perfect fifth up/down (following the circle of fifths)
"""

from europi import *
from europi_script import EuroPiScript

from experimental.quantizer import CommonScales, SEMITONE_LABELS, VOLTS_PER_OCTAVE, VOLTS_PER_SEMITONE, SEMITONES_PER_OCTAVE
from experimental.random_extras import shuffle

import random

MODE_ASCENDING = 0
MODE_DESCENDING = 1
MODE_RANDOM = 2


class Arpeggio:
    def __init__(self, scale, mode):
        """Create an arpeggio using the notes from the given scale

        @param scale  A quantizer scales whose notes will be used
        @param mode   One of ascending, descending, or random
        """
        self.mode = mode
        self.change_scale(scale)

    def change_scale(self, scale):
        """Change the current scale

        @param scale  The new quantizer scale we want to use
        """
        self.semitones = [
            i for i in range(len(scale.notes)) if scale[i]
        ]
        if self.mode == MODE_RANDOM:
            # shuffle so we have a chance to choose the last note first
            # see @next_node
            shuffle(self.semitones)

    def next_note(self):
        """Get the next note that should be played

        @return The semitone as an integer from 0-12
        """
        if self.mode == MODE_ASCENDING:
            semitone = self.semitones.pop(0)
            self.semitones.append(semitone)
        elif self.mode == MODE_DESCENDING:
            semitone = self.semitones.pop(-1)
            self.semitones.insert(0, semitone)
        else:
            # never choose the _last_ note in the array
            # move the chosen note to the end of the list to avoid repeats
            n = random.randint(0, len(self.semitones)-2)
            semitone = self.semitones.pop(n)
            self.semitones.append(n)

        return semitone

    def __len__(self):
        return len(self.semitones)


class Arpeggiator(EuroPiScript):
    def __init__(self):
        super().__init__()

        ## Indicates if the GUI needs to be refreshed
        self.ui_dirty = True

        ## The available scales to be played
        self.scales = [
            CommonScales.Chromatic,

            CommonScales.NatMajor,
            CommonScales.HarMajor,
            CommonScales.Major135,
            CommonScales.Major1356,
            CommonScales.Major1357,

            CommonScales.NatMinor,
            CommonScales.HarMinor,
            CommonScales.Minor135,
            CommonScales.Minor1356,
            CommonScales.Minor1357,

            CommonScales.MajorBlues,
            CommonScales.MinorBlues,

            CommonScales.WholeTone,
            CommonScales.Pentatonic,
            CommonScales.Dominant7
        ]

        ## The active scale within @self.scales
        self.current_scale_index = 0

        ## Have we received an external trigger to be processed?
        self.trigger_recvd = False

        ## Should we change the scale on the next trigger?
        self.scale_changed = False

        ## What octave range are we playing (minimum 1)
        self.n_octaves = 1

        ## What is the current octave we're playing (range [0, n_octaves))
        self.current_octave = 0

        ## What is the root note of our scale (semitone, default C=0)
        self.root = 0

        ## What is the lowest octave we play
        self.root_octave = 0

        @din.handler
        def on_din():
            self.trigger_recvd = True

        @b1.handler
        def on_b1_press():
            self.current_scale_index = (self.current_scale_index - 1) % len(self.scales)
            self.scale_changed = True
            self.ui_dirty = True

        @b2.handler
        def on_b2_press():
            self.current_scale_index = (self.current_scale_index + 1) % len(self.scales)
            self.scale_changed = True
            self.ui_dirty = True

        self.load()
        self.set_scale()

    def save(self):
        state = {
            "scale": self.current_scale_index
        }
        self.save_state_json(state)

    def load(self):
        settings = self.load_state_json()
        self.current_scale_index = settings.get("scale", 0)

    def set_scale(self):
        # keep track of the number of notes we've played; once we play enough we may need to
        # go up/down an octave
        self.n_notes_played = 0

        # the current root octave for ascending arpeggios
        self.current_octave = 0

        scale = self.scales[self.current_scale_index]
        self.arps = [
            Arpeggio(scale, MODE_ASCENDING),  # ascending notes, ascending octaves
            Arpeggio(scale, MODE_ASCENDING),  # ascending notes, descending octaves
            Arpeggio(scale, MODE_RANDOM),     # random notes, ascending octaves
            Arpeggio(scale, MODE_DESCENDING), # descending notes, ascending octaves
            Arpeggio(scale, MODE_DESCENDING), # descending notes, descending octaves
            Arpeggio(scale, MODE_RANDOM)      # random notes, random octaves
        ]

    def tick(self):
        # apply the new scale in-sync with the incoming triggers so we don't get out-of-phase changes
        if self.scale_changed:
            self.scale_changed = False
            self.set_scale()
            self.save()

        # apply the output voltages; each one is slightly unique

        # CV1: ascending arpeggio, ascending octaves
        volts = (self.root_octave + self.current_octave) * VOLTS_PER_OCTAVE + (self.root + self.arps[0].next_note()) * VOLTS_PER_SEMITONE
        cv1.voltage(volts)

        # CV2: ascending arpeggio, descending octaves
        volts = (self.root_octave + self.n_octaves - self.current_octave - 1) * VOLTS_PER_OCTAVE + (self.root + self.arps[1].next_note()) * VOLTS_PER_SEMITONE
        cv2.voltage(volts)

        # CV3: random arpeggio, ascending octaves
        volts = (self.root_octave + self.current_octave) * VOLTS_PER_OCTAVE + (self.root + self.arps[2].next_note()) * VOLTS_PER_SEMITONE
        cv3.voltage(volts)

        # CV4: descending arpeggio, ascending octaves
        volts = (self.root_octave + self.current_octave) * VOLTS_PER_OCTAVE + (self.root + self.arps[3].next_note()) * VOLTS_PER_SEMITONE
        cv4.voltage(volts)

        # CV5: descending arpeggio, descending octaves
        volts = (self.root_octave + self.n_octaves - self.current_octave - 1) * VOLTS_PER_OCTAVE + (self.root + self.arps[4].next_note()) * VOLTS_PER_SEMITONE
        cv5.voltage(volts)

        # CV6: random arpeggio, random octave
        volts = random.randint(self.root_octave, self.root_octave + self.n_octaves) * VOLTS_PER_OCTAVE + (self.root + self.arps[5].next_note()) * VOLTS_PER_SEMITONE
        cv6.voltage(volts)

        # Increment the note & octave counter
        self.n_notes_played += 1
        if self.n_notes_played >= len(self.arps[0]):
            self.current_octave += 1
            self.n_notes_played = 0
            if self.current_octave >= self.n_octaves:
                self.current_octave = 0

    def main(self):
        prev_root_octave = self.root_octave
        prev_n_octaves = self.n_octaves
        prev_root = self.root

        while True:
            # Update the desired root, octave, and octave range according to the analogue inputs
            self.root_octave = int(k1.percent() * 5)
            self.n_octaves = int(k2.percent() * 5) + 1
            self.root = int(ain.read_voltage() / VOLTS_PER_SEMITONE) % SEMITONES_PER_OCTAVE

            # Update the CV outputs if we've received a clock signal
            if self.trigger_recvd:
                self.trigger_recvd = False
                self.tick()

            # Re-render the OLED if needed
            # This should only occur if we've pressed a button or the analogue inputs have changed
            self.ui_dirty = (self.ui_dirty or
                prev_root_octave != self.root_octave or
                prev_n_octaves != self.n_octaves or
                prev_root != self.root
            )
            if self.ui_dirty:
                self.ui_dirty = False
                oled.fill(0)
                # OLED displays something like:
                # +--------------+
                # |    F#1-3     |   <- Root note, root octave, highest octave (adjust: AIN, K1, K2)
                # |   Min 1356   |   <- Current scale/arpeggio selection (adjust: B1/B2)
                # +--------------+
                oled.centre_text(f"""{SEMITONE_LABELS[self.root]}{self.root_octave}-{self.root_octave + self.n_octaves - 1}
{self.scales[self.current_scale_index]}""")
                oled.show()

            # Keep track of previous analogue settings so we can update the GUI again if needed
            prev_root_octave = self.root_octave
            prev_n_octaves = self.n_octaves
            prev_root = self.root


if __name__ == "__main__":
    Arpeggiator().main()



# Copyright 2024 Allen Synthesis
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""See menu.md for details."""
# Reset the module state and display bootsplash screen.
from europi import bootsplash


bootsplash()


from bootloader import BootloaderMenu
from collections import OrderedDict

## Scripts that are included in the menu
#
#  Keys are the names displayed in the menu, values are the fully-qualified names
#  of the classes to launch.  The classes MUST be EuriPiScript subclasses
#
#  The OLED can display up to 16 characters horizontally, so make sure the names fit
#  that width requirement
#
# fmt: off
EUROPI_SCRIPTS = OrderedDict([
#   ["0123456789abcdef",  "contrib.spam.Eggs"],
    ["Arpeggiator",       "contrib.arp.Arpeggiator"],
    ["Bernoulli Gates",   "contrib.bernoulli_gates.BernoulliGates"],
    ["Bezier Curves",     "contrib.bezier.Bezier"],
    ["Binary Counter",    "contrib.binary_counter.BinaryCounter"],
    ["Bit Garden",        "contrib.bit_garden.BitGarden"],
    ["Clock Modifier",    "contrib.clock_mod.ClockModifier"],
    ["Coin Toss",         "contrib.coin_toss.CoinToss"],
    ["Consequencer",      "contrib.consequencer.Consequencer"],
    ["Conway",            "contrib.conway.Conway"],
    ["CVecorder",         "contrib.cvecorder.CVecorder"],
    ["Daily Random",      "contrib.daily_random.DailyRandom"],
    ["DCSN-2",            "contrib.dscn2.Dcsn2"],
    ["DFAM Controller",   "contrib.dfam.DfamController"],
    ["EgressusMelodiam",  "contrib.egressus_melodiam.EgressusMelodiam"],
    ["EnvelopeGen",       "contrib.envelope_generator.EnvelopeGenerator"],
    ["Euclid",            "contrib.euclid.EuclideanRhythms"],
    ["Gates & Triggers",  "contrib.gates_and_triggers.GatesAndTriggers"],
    ["Gate Phaser",       "contrib.gate_phaser.GatePhaser"],
    ["Hamlet",            "contrib.hamlet.Hamlet"],
    ["HarmonicLFOs",      "contrib.harmonic_lfos.HarmonicLFOs"],
    ["HTTP Interface",    "contrib.http_control.HttpControl"],
    ["Itty Bitty",        "contrib.itty_bitty.IttyBitty"],
    ["Kompari",           "contrib.kompari.Kompari"],
    ["Logic",             "contrib.logic.Logic"],
    ["Lutra",             "contrib.lutra.Lutra"],
    ["MasterClock",       "contrib.master_clock.MasterClock"],
    ["Morse",             "contrib.morse.Morse"],
    ["NoddyHolder",       "contrib.noddy_holder.NoddyHolder"],
    ["OSC Interface",     "contrib.osc_control.OscControl"],
    ["Pam's Workout",     "contrib.pams.PamsWorkout2"],
    ["Particle Phys.",    "contrib.particle_physics.ParticlePhysics"],
    ["Pet Rock",          "contrib.pet_rock.PetRock"],
    ["Piconacci",         "contrib.piconacci.Piconacci"],
    ["Poly Square",       "contrib.poly_square.PolySquare"],
    ["PolyrhythmSeq",     "contrib.polyrhythmic_sequencer.PolyrhythmSeq"],
    ["Probapoly",         "contrib.probapoly.Probapoly"],
    ["Quantizer",         "contrib.quantizer.QuantizerScript"],
    ["RadioScanner",      "contrib.radio_scanner.RadioScanner"],
    ["Scope",             "contrib.scope.Scope"],
    ["Seq. Switch",       "contrib.sequential_switch.SequentialSwitch"],
    ["Sigma",             "contrib.sigma.Sigma"],
    ["Slopes",            "contrib.slopes.Slopes"],
    ["Smooth Rnd Volts",  "contrib.smooth_random_voltages.SmoothRandomVoltages"],
    ["StrangeAttractor",  "contrib.strange_attractor.StrangeAttractor"],
    ["Traffic",           "contrib.traffic.Traffic"],
    ["Turing Machine",    "contrib.turing_machine.EuroPiTuringMachine"],
    ["Volts",             "contrib.volts.OffsetVoltages"],

    # Examples & proof-of-concept scripts that aren't generally useful
    # but someone might want to enable to test out
    #["Custom Fonts",      "contrib.custom_font_demo.CustomFontDemo"],
    #["HelloWorld",        "contrib.hello_world.HelloWorld"],
    #["KnobPlayground",    "contrib.knob_playground.KnobPlayground"],
    #["Menu Example",      "contrib.settings_menu_example.SettingsMenuExample"],


    # System tools, in alphabetical order with a _ prefix

    ["_About",            "tools.about.About"],
    ["_BootloaderMode",   "bootloader_mode.BootloaderMode"],
    ["_Calibrate",        "tools.calibrate.Calibrate"],
    ["_Config Editor",    "tools.conf_edit.ConfigurationEditor"],
    ["_Diagnostic",       "tools.diagnostic.Diagnostic"],
    ["_Exp Cfg Editor",   "tools.experimental_conf_edit.ExperimentalConfigurationEditor"],
])
# fmt: on

if __name__ == "__main__":
    BootloaderMenu(EUROPI_SCRIPTS).main()



# Copyright 2024 Allen Synthesis
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from setuptools import setup

__version__ = ""
exec(open("./firmware/version.py").read())

setup(
    name="micropython-europi-contrib",
    version=__version__,
    description="Community Contributions to EuroPi module for MicroPython",
    long_description="Community Contributions to EuroPi module for MicroPython",
    url="https://github.com/Allen-Synthesis/EuroPi",
    author="Allen Synthesis",
    author_email="contact@allensynthesis.co.uk",
    license="Apache 2.0",
    packages=["contrib"],
    data_files=[("/", ["main.py"])],
    py_modules=["firmware.version"],
    namespace_packages=["contrib"],
)



# Copyright 2024 Allen Synthesis
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from time import sleep
from europi import oled, b1, din
from europi_script import EuroPiScript


class HelloWorld(EuroPiScript):
    def __init__(self):
        super().__init__()
        state = self.load_state_json()

        self.counter = state.get("counter", 0)
        self.enabled = state.get("enabled", True)

        din.handler(self.increment_counter)
        b1.handler(self.toggle_enablement)

    @classmethod
    def display_name(cls):
        return "Hello World"

    def increment_counter(self):
        if self.enabled:
            self.counter += 1
            self.save_state()

    def toggle_enablement(self):
            self.enabled = not self.enabled
            self.save_state()

    def save_state(self):
        """Save the current state variables as JSON."""
        # Don't save if it has been less than 5 seconds since last save.
        if self.last_saved() < 5000:
            return

        state = {
            "counter": self.counter,
            "enabled": self.enabled,
        }
        self.save_state_json(state)

    def main(self):
        while True:
            oled.centre_text(f"Hello world\n{self.counter}")
            sleep(1)

if __name__ == "__main__":
    HelloWorld().main()



#!/usr/bin/env python3
# Copyright 2024 Allen Synthesis
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Clock multiplier & divider
"""

import time

from collections import OrderedDict
from europi import *
from europi_script import EuroPiScript
from experimental.a_to_d import AnalogReaderDigitalWrapper
from experimental.knobs import KnobBank
from experimental.screensaver import Screensaver
from math import floor


# This script operates in microseconds, so for convenience define one second as a constant
ONE_SECOND = 1000000

# Automatically reset if we receive no input clocks after 5 seconds
# This lets us handle resonably slow input clocks, while also providing some reasonable timing
CLOCK_IN_TIMEOUT = 5 * ONE_SECOND


def ljust(s, length):
    """Re-implements the str.ljust method from standard Python

    @param s  A string to left-align with spaces
    @param length  The desired final length of the string. If len(s) is greater or equal to this, no padding is done
    @return The padding string
    """
    n_spaces = max(0, length - len(s))
    return s + ' '*n_spaces


class ClockOutput:
    """A control class that handles a single output
    """
    ## The smallest common multiple of the allowed clock divisions (2, 3, 4, 5, 6, 8, 12)
    #
    #  Used to reset the input gate counter to avoid integer overflows/performance degredation with large values
    MAX_GATE_COUNT = 120

    def __init__(self, output_port, modifier):
        """Constructor

        @param output_port  One of the six output CV ports, e.g. cv1, cv2, etc... that this class will control
        @param modifier     The initial clock modifier for this output channel
        """
        self.last_external_clock_at = time.ticks_us()
        self.last_interval_us = 0
        self.modifier = modifier
        self.output_port = output_port

        # Should the output be high or low?
        self.is_high = False

        # Used to implement basic gate-skipping for clock divisions
        self.input_gate_counter = 0

    def set_external_clock(self, ticks_us):
        """Notify this output when the last external clock signal was received.

        The calculate_state function will use this to calculate the duration of the high/low phases
        of the output gate
        """
        if self.last_external_clock_at != ticks_us:
            self.last_interval_us = time.ticks_diff(ticks_us, self.last_external_clock_at)
            self.last_external_clock_at = ticks_us

            self.input_gate_counter += 1
            if self.input_gate_counter >= self.MAX_GATE_COUNT:
                self.input_gate_counter = 0

    def calculate_state(self, ticks_us):
        """Calculate whether this output should be high or low based on the current time

        Must be called before calling set_output_voltage

        @param ticks_us  The current time in microseconds; passed as a parameter to synchronize multiple channels
        """
        if self.modifier >= 1:
            # We're in clock multiplication mode; calculate the duration of output gates and set high/low state
            gate_duration_us = self.last_interval_us / self.modifier
            hi_lo_duration_us = gate_duration_us / 2

            # The time elapsed since our last external clock
            elapsed_us = time.ticks_diff(ticks_us, self.last_external_clock_at)

            # The number of phases that have happened since the last incoming clock
            n_phases = elapsed_us // hi_lo_duration_us

            self.is_high = n_phases % 2 == 0
        else:
            # We're in clock division mode; just do a simple gate-skip to stay in sync with the input
            n_gates = round(1.0 / self.modifier)
            self.is_high = self.input_gate_counter % (n_gates * 2) < n_gates

    def set_output_voltage(self):
        """Set the output voltage either high or low.

        Must be called after calling calculate_state
        """
        if self.is_high:
            self.output_port.on()
        else:
            self.output_port.off()

    def reset(self):
        """Reset the pattern to the initial position
        """
        self.is_high = False
        self.output_port.off()
        self.input_gate_counter = 0


class ClockModifier(EuroPiScript):
    """The main script class; multiplies and divides incoming clock signals
    """
    def __init__(self):
        self.ui_dirty = False
        state = self.load_state_json()

        self.k1_bank = (
            KnobBank.builder(k1) \
            .with_unlocked_knob("channel_a") \
            .with_locked_knob("channel_b", initial_percentage_value=state.get("mod_cv2", 0.5)) \
            .with_locked_knob("channel_c", initial_percentage_value=state.get("mod_cv3", 0.5)) \
            .build()
        )

        self.k2_bank = (
            KnobBank.builder(k2) \
            .with_unlocked_knob("channel_a") \
            .with_locked_knob("channel_b", initial_percentage_value=state.get("mod_cv5", 0.5)) \
            .with_locked_knob("channel_c", initial_percentage_value=state.get("mod_cv6", 0.5)) \
            .build()
        )

        ## The time the last rising edge of the clock was recorded
        #
        #  Initially negative 1s to avoid starting the module
        self.last_clock_at = -1000

        ## The output channels
        self.outputs = [
            ClockOutput(cv, 1.0) for cv in cvs
        ]

        ## Gui labels to indicate what row of modifiers we're adjusting
        self.channel_markers = ['>', ' ', ' ']

        self.clock_modifiers = OrderedDict([
            ["/12", 1.0/12.0],
            ["/8", 1.0/8.0],
            ["/6", 1.0/6.0],
            ["/5", 1.0/5.0],
            ["/4", 1.0/4.0],
            ["/3", 1.0/3.0],
            ["/2", 1.0/2.0],
            ["x1", 1.0],
            ["x2", 2.0],
            ["x3", 3.0],
            ["x4", 4.0],
            ["x5", 5.0],
            ["x6", 6.0],
            ["x8", 8.0],
            ["x12", 12.0]
        ])

        # Indicates that the modifiers have changed and a save is required
        self.state_dirty = False

        @b1.handler
        def b1_rising():
            """Activate channel b controls while b1 is held
            """
            self.k1_bank.set_current("channel_b")
            self.k2_bank.set_current("channel_b")
            self.channel_markers[0] = ' '
            self.channel_markers[1] = '>'
            self.channel_markers[2] = ' '
            self.ui_dirty = True

        @b2.handler
        def b2_rising():
            """Activate channel c controls while b1 is held
            """
            self.k1_bank.set_current("channel_c")
            self.k2_bank.set_current("channel_c")
            self.channel_markers[0] = ' '
            self.channel_markers[1] = ' '
            self.channel_markers[2] = '>'
            self.ui_dirty = True

        @b1.handler_falling
        def b1_falling():
            """Revert to channel a when the button is released
            """
            self.k1_bank.set_current("channel_a")
            self.k2_bank.set_current("channel_a")
            self.channel_markers[0] = '>'
            self.channel_markers[1] = ' '
            self.channel_markers[2] = ' '
            self.state_dirty = True
            self.ui_dirty = True

        @b2.handler_falling
        def b2_falling():
            """Revert to channel a when the button is released
            """
            self.k1_bank.set_current("channel_a")
            self.k2_bank.set_current("channel_a")
            self.channel_markers[0] = '>'
            self.channel_markers[1] = ' '
            self.channel_markers[2] = ' '
            self.state_dirty = True
            self.ui_dirty = True

        @din.handler
        def on_din():
            """Record the start time of our rising edge
            """
            self.last_clock_at = time.ticks_us()

        def on_ain():
            """Reset all channels when AIN goes high
            """
            for output in self.outputs:
                output.reset()

        self.d_ain = AnalogReaderDigitalWrapper(
            ain,
            cb_rising=on_ain
        )

    def save_state(self):
        """Save the clock modifiers for channels 2, 3, 5, 6 to the config file for loading

        Channels 1 and 4 are read directly from the knob positions on startup, and are considered volatile
        """
        state = {
            "mod_cv2": self.k1_bank["channel_b"].percent(),
            "mod_cv3": self.k1_bank["channel_c"].percent(),
            "mod_cv5": self.k2_bank["channel_b"].percent(),
            "mod_cv6": self.k2_bank["channel_c"].percent()
        }
        self.save_state_json(state)

    def main(self):
        """The main loop
        """
        screensaver = Screensaver()
        last_render_at = time.ticks_us()

        knob_choices = list(self.clock_modifiers.keys())

        prev_mods = [
            knob_choices[0],
            knob_choices[0],
            knob_choices[0],
            knob_choices[0],
            knob_choices[0],
            knob_choices[0],
        ]

        while True:
            # update AIN so its rising edge callback can fire
            self.d_ain.update()

            # Save the clock modifiers for channels 2, 3, 5, 6 if they've been edited
            if self.state_dirty:
                self.state_dirty = False
                self.save_state()

            # Get the clock modifiers for all 6 channels and apply them to the outputs
            mods = [
                self.k1_bank["channel_a"].choice(knob_choices),
                self.k1_bank["channel_b"].choice(knob_choices),
                self.k1_bank["channel_c"].choice(knob_choices),
                self.k2_bank["channel_a"].choice(knob_choices),
                self.k2_bank["channel_b"].choice(knob_choices),
                self.k2_bank["channel_c"].choice(knob_choices)
            ]
            for i in range(len(mods)):
                self.outputs[i].modifier = self.clock_modifiers[mods[i]]

            # if we don't get an external signal within the timeout duration, reset the outputs
            now = time.ticks_us()
            if time.ticks_diff(now, self.last_clock_at) <= CLOCK_IN_TIMEOUT:
                # separate calculating the high/low state and setting the output voltage into two loops
                # this helps reduce phase-shifting across outputs
                for output in self.outputs:
                    output.set_external_clock(self.last_clock_at)
                    output.calculate_state(now)

                for output in self.outputs:
                    output.set_output_voltage()
            else:
                for output in self.outputs:
                    output.reset()

            # Update the GUI
            # This only needs to be done if the modifiers have changed or a button has been pressed/released
            self.ui_dirty = self.ui_dirty or any([mods[i] != prev_mods[i] for i in range(len(mods))])
            if self.ui_dirty:
                # Yes, this is a very long string, but it centers nicely
                # It looks something like this:
                #
                #    > 1: x1   4: /2
                #      2: x2   5: x3
                #      3: /4   6: /3
                #
                oled.fill(0)
                oled.centre_text(
                    f"{self.channel_markers[0]} 1:{ljust(mods[0], 3)} 4:{ljust(mods[3], 3)}\n{self.channel_markers[1]} 2:{ljust(mods[1], 3)} 5:{ljust(mods[4], 3)}\n{self.channel_markers[2]} 3:{ljust(mods[2], 3)} 6:{ljust(mods[5], 3)}"
                )
                self.ui_dirty = False
                last_render_at = time.ticks_us()
            elif time.ticks_diff(now, last_render_at) > screensaver.ACTIVATE_TIMEOUT_US:
                last_render_at = time.ticks_add(now, -screensaver.ACTIVATE_TIMEOUT_US)
                screensaver.draw()

            for i in range(len(mods)):
                prev_mods[i] = mods[i]


if __name__=="__main__":
    ClockModifier().main()



# Copyright 2022 Allen Synthesis
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from europi import *
import machine
from time import ticks_diff, ticks_ms
from random import randint, uniform, choice

from europi_script import EuroPiScript

"""
Piconacci
author: Sean Bechhofer (github.com/seanbechhofer)
date: 2022-05
labels: triggers, maths

Triggers based on the Fibonacci sequence

digital_in: clock in
analog_in:

knob_1:
knob_2:

button_1: Short Press: Move window on sequence left. Long Press: Rotate values left
button_2: Short Press: Move window on sequence to the right. Long Press: Rotate values right
output_1: trigger
output_2: trigger
output_3: trigger
output_4: trigger
output_5: trigger
output_6: trigger

"""


# This probably works for most sensible circumstances. The 50th value
# (disregarding 0, 1) is 20,365,011,074. At a BPM of 120, that'll be a
# trigger every 300 years which is quite a lot even for slow moving
# ambient pieces,
MAX_FIB = 50


class Piconacci(EuroPiScript):
    def __init__(self):
        # Flag to indicate whether the display needs updating.
        self.display_update_required = True
        # offset determines the sublist to be used for the values.
        self.offset = 0
        # rotate adjusts how the values are mapped to the cv
        # outputs. A value of 0 means that cv1 is controlled by the
        # first, cv2 the second etc. A rotate value of 1 means that
        # cv2 is controlled by the first, cv3 the second, and cv1 the
        # last.
        self.rotate = 0
        # Overclock the Pico for improved performance.
        machine.freq(250_000_000)
        # List of fibonacci numbers.
        self.fib = [1, 1]
        for i in range(0, MAX_FIB):
            self.fib.append(self.fib[-2] + self.fib[-1])
        # Strip off first 1
        self.fib = self.fib[1:]

        self.steps = [0] * 6

        @b1.handler_falling
        def b1Pressed():
            if ticks_diff(ticks_ms(), b1.last_pressed()) > 300:
                # Rotate values left
                self.rotate = (self.rotate - 1) % 6
            else:
                # Shift all the values left in the sequence
                if self.offset > 0:
                    self.offset -= 1
            # State has changed
            self.display_update_required = True

        @b2.handler_falling
        def b2Pressed():
            if ticks_diff(ticks_ms(), b2.last_pressed()) > 300:
                # Rotate values right
                self.rotate = (self.rotate + 1) % 6
            else:
                # Shift all the values right in the sequence
                if self.offset < MAX_FIB - 7:
                    self.offset += 1
            # State has changed
            self.display_update_required = True

        # Triggered on each clock into digital input. Output triggers.
        @din.handler
        def clockTrigger():
            for tr in range(0, 6):
                self.steps[tr] += 1
                if self.steps[tr] >= self.value(tr):
                    # Trigger on tr
                    cvs[tr].on()
                    # If the values change due to shift or rotation,
                    # we may get slightly odd behaviour here,
                    # particularly if the values reduce.
                    self.steps[tr] = 0

        @din.handler_falling
        def clockTriggerEnd():
            for cv in cvs:
                cv.off()

    def value(self, index):
        # Return a value from the series, taking into account offset and rotation
        return self.fib[self.offset : self.offset + 6][(index + self.rotate) % 6]

    def main(self):
        # Reset all outputs
        turn_off_all_cvs()
        while True:
            # If the state has changed, update display
            if self.display_update_required:
                self.updateScreen()
                self.display_update_required = False

    def updateScreen(self):
        oled.fill(0)

        # Show the values.
        oled.text("Piconacci", 28, 0, 1)
        oled.text(str(self.value(0)), 10, 12, 1)
        oled.text(str(self.value(1)), 50, 12, 1)
        oled.text(str(self.value(2)), 90, 12, 1)
        oled.text(str(self.value(3)), 10, 24, 1)
        oled.text(str(self.value(4)), 50, 24, 1)
        oled.text(str(self.value(5)), 90, 24, 1)
        oled.show()


if __name__ == "__main__":
    pc = Piconacci()
    pc.main()



# Copyright 2024 Allen Synthesis
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from europi import *
from time import ticks_diff, ticks_ms
#from random import uniform
from europi_script import EuroPiScript
#from europi_config import EuroPiConfig

"""
Gate Phaser
author: Nik Ansell (github.com/gamecat69)
date: May 2024
labels: sequencer, gates
"""

# Constants
KNOB_CHANGE_TOLERANCE = 0.001
MIN_CYCLE_TIME_MS = 100
MIN_PHASE_SHIFT_MS = 5
MIN_MS_BETWEEN_SAVES = 2000
GATE_LENGTH_MS = 20

class GatePhaser(EuroPiScript):
    def __init__(self):

        # Initialize variables

        # How many multiples of the gate delay time will each gate be delayed?
        self.gateDelayMultiples = [ [0,1,2,3,4,5],[2,3,4,5,6,7],[0,1,2,3,6,9],[1,2,3,2,4,6],[5,4,3,2,1,0] ]
        # UI only, changes the behaviour of the gate delay control
        self.gateDelayControlOptions = [5, 10, 20]
        # Lists containing params for each output
        self.gateDelays = []
        self.gateOnTimes = []
        self.gateOffTimes = []
        self.gateStates = []

        self.lastK1Reading = 0
        self.lastK2Reading = 0
        self.lastSaveState = ticks_ms()
        self.pendingSaveState = False
        self.screenRefreshNeeded = True

        self.smoothK1 = 0
        self.smoothK2 = 0
        self.loadState()

        # Populate working lists
        self.calcGateDelays(newList=True)
        self.calcGateTimes(newList=True)

        # Create intervalStr for the UI
        self.buildIntervalStr()

        # -----------------------------
        # Interupt Handling functions
        # -----------------------------

        @din.handler
        def resetGates():
            """Resets gate timers"""
            self.calcGateDelays()
            self.calcGateTimes()

        @b1.handler_falling
        def b1Pressed():
            """Triggered when B1 is pressed and released. Select gate delay multiples"""
            self.selectedGateDelayMultiple = (self.selectedGateDelayMultiple + 1) % len(self.gateDelayMultiples)
            self.calcGateDelays()
            self.calcGateTimes()
            self.buildIntervalStr()
            self.screenRefreshNeeded = True
            self.pendingSaveState = True


        @b2.handler_falling
        def b2Pressed():
            """Triggered when B2 is pressed and released. Select gate control multiplier"""
            self.selectedGateControlMultiplier = (self.selectedGateControlMultiplier + 1) % len(self.gateDelayControlOptions)
            self.calcGateDelays()
            self.calcGateTimes()
            self.screenRefreshNeeded = True
            self.pendingSaveState = True


    def buildIntervalStr(self):
        """Create a string for the UI showing the gate delay multiples"""
        self.intervalsStr = ''
        for i in self.gateDelayMultiples[self.selectedGateDelayMultiple]:
            self.intervalsStr = self.intervalsStr + str(i) + ':'


    def lowPassFilter(self, alpha, prevVal, newVal):
        """Smooth out some analogue noise. Higher Alpha = more smoothing"""
        # Alpha value should be between 0 and 1.0
        return alpha * prevVal + (1 - alpha) * newVal

    def calcGateDelays(self, newList=False):
        """Populate a list containing the gate delay in ms for each output"""
        for n in range(6):
            val = self.gateDelayMultiples[self.selectedGateDelayMultiple][n] * self.slaveGateIntervalMs
            if newList:
                self.gateDelays.append(val)
            else:
                self.gateDelays[n] = (val)


    def calcGateTimes(self, newList=False):
        """Calculate the next gate on and off times based on the current time"""
        self.currentTimeStampMs = ticks_ms()
        for n in range(6):
            gateOnTime = self.currentTimeStampMs + self.gateDelays[n]
            gateOffTime = gateOnTime + GATE_LENGTH_MS
            if newList:
                self.gateOnTimes.append(gateOnTime)
                self.gateOffTimes.append(gateOffTime)
                self.gateStates.append(False)
            else:
                self.gateOnTimes[n] = gateOnTime
                self.gateOffTimes[n] = gateOffTime
                self.gateStates[n] = False


    def getKnobValues(self):
        """Get k1 and k2 values and adjust working parameters if knobs have moved"""
        changed = False

        # Get knob values and smooth using a simple low pass filter
        self.smoothK1 = int(self.lowPassFilter(0.15, self.lastK1Reading, k1.read_position(100) + 2))
        self.smoothK2 = int(self.lowPassFilter(0.15, self.lastK2Reading, k2.read_position(100) + 2))

        if abs(self.smoothK1 - self.lastK1Reading) > KNOB_CHANGE_TOLERANCE:
            self.masterGateIntervalMs = max(MIN_CYCLE_TIME_MS, self.smoothK1 * 25)
            changed = True

        if abs(self.smoothK2 - self.lastK2Reading) > KNOB_CHANGE_TOLERANCE:
            self.slaveGateIntervalMs = max(MIN_PHASE_SHIFT_MS, self.smoothK2 * self.gateDelayControlOptions[self.selectedGateControlMultiplier])
            changed = True

        if changed:
            self.calcGateDelays()
            self.calcGateTimes()
            self.screenRefreshNeeded = True
            self.pendingSaveState = True

        self.lastK1Reading = self.smoothK1
        self.lastK2Reading = self.smoothK2

    def main(self):
        """Entry point - main loop. See inline comments for more info"""
        while True:
            self.getKnobValues()
            if self.screenRefreshNeeded:
                self.updateScreen()

            # Cycle through outputs turning gates on and off as needed
            # When a gate is triggered it's next on and off time is calculated
            self.currentTimeStampMs = ticks_ms()
            for n in range(len(cvs)):
                if self.currentTimeStampMs >= self.gateOffTimes[n] and self.gateStates[n]:
                    cvs[n].off()
                    self.gateStates[n] = False
                elif self.currentTimeStampMs >= self.gateOnTimes[n] and not self.gateStates[n]:
                    cvs[n].on()
                    self.gateStates[n] = True
                    # When will the gate need to turn off?
                    self.gateOffTimes[n] = self.currentTimeStampMs + GATE_LENGTH_MS
                    # When will the next gate need to fire?
                    self.gateOnTimes[n] = self.currentTimeStampMs + self.gateDelays[n] + self.masterGateIntervalMs

            # Save state
            if self.pendingSaveState and ticks_diff(ticks_ms(), self.lastSaveState) >= MIN_MS_BETWEEN_SAVES:
                self.saveState()
                self.pendingSaveState = False

    def updateScreen(self):
        """Update the screen only if something has changed. oled.show() hogs the processor and causes latency."""

        # Clear screen
        oled.fill(0)

        oled.text("Cycle", 5, 0, 1)
        oled.text(str(self.masterGateIntervalMs), 5, 10, 1)
        oled.text("Delay", 80, 0, 1)
        oled.text(str(self.slaveGateIntervalMs), 80, 10, 1)
        oled.text(self.intervalsStr[:-1], 0, 22, 1)
        oled.text('x' + str(self.gateDelayControlOptions[self.selectedGateControlMultiplier]), 104, 22, 1)

        oled.show()
        self.screenRefreshNeeded = False

    def saveState(self):
        """Save working vars to a save state file"""
        self.state = {
            "masterGateIntervalMs": self.masterGateIntervalMs,
            "slaveGateIntervalMs": self.slaveGateIntervalMs,
            "selectedGateDelayMultiple": self.selectedGateDelayMultiple,
            "selectedGateControlMultiplier": self.selectedGateControlMultiplier,
        }
        self.save_state_json(self.state)
        self.lastSaveState = ticks_ms()

    def loadState(self):
        """Load a previously saved state, or initialize working vars, then save"""
        self.state = self.load_state_json()
        self.masterGateIntervalMs = self.state.get("masterGateIntervalMs", 1000)
        self.slaveGateIntervalMs = self.state.get("slaveGateIntervalMs", 100)
        self.selectedGateDelayMultiple = self.state.get("selectedGateDelayMultiple", 0)
        self.selectedGateControlMultiplier = self.state.get("selectedGateControlMultiplier", 0)

if __name__ == "__main__":
    GatePhaser().main()



# Copyright 2024 Allen Synthesis
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from europi import *
from math import cos, radians
from time import sleep_ms, ticks_diff, ticks_ms
from machine import freq
from random import randint
from europi_script import EuroPiScript

MAX_VOLTAGE = MAX_OUTPUT_VOLTAGE # Default is inherited but this can be overriden by replacing "MAX_OUTPUT_VOLTAGE" with an integer
MAX_HARMONIC = 32 # Too high a value may be hard to select using the knob, but the actual hardware limit is only reached at 4096
LONG_PRESS_MIN_DURATION = 500

class HarmonicLFOs(EuroPiScript):
    MODES_SHAPES = {
        'SINE': 0,
        'SAW': 1,
        'SQUARE': 2,
        'OFF': 3,
        'RANDOM': 4,
        'NOISE': 5,
    }

    def __init__(self):
        super().__init__()

        # Retreive saved state information from file
        state = self.load_state_json()

        # Use the saved values for the LFO divisions and mode if found in the save state file, using defaults if not
        self.divisions = state.get("divisions", [1, 3, 5, 7, 11, 13])
        self.modes = state.get("modes", [self.MODES_SHAPES['SINE']] * 6)
        self.MODES_COUNT = len(self.MODES_SHAPES)
        self.viewAllWaveforms = state.get("viewAllWaveforms", True)

        # Initialise all the other variables
        self.degree = 0
        self.rad = radians(self.degree)
        self.delay, self.increment_value = self.get_delay_increment_value()
        self.pixel_x = OLED_WIDTH-1
        self.pixel_y = OLED_HEIGHT-1
        self.selected_lfo = 0
        self.clock_division = self.selected_lfo_start_value = self.get_clock_division()
        self.state_needs_saving = False

        # Set the digital input and button handlers
        din.handler(self.reset)

        @b1.handler_falling
        def b1Pressed():
            """Triggered when B1 is pressed"""
            diff = ticks_diff(ticks_ms(), b1.last_pressed())
            if diff > LONG_PRESS_MIN_DURATION:
                """Long press: toggle waveform view mode"""
                self.viewAllWaveforms = not self.viewAllWaveforms
            else:
                """Short press: Change the mode that controls wave shape"""
                self.modes[self.selected_lfo] = (self.modes[self.selected_lfo] + 1) % self.MODES_COUNT

            self.state_needs_saving = True

        @b2.handler_falling
        def b2Pressed():
            """Move the selection to the next LFO"""
            self.selected_lfo = (self.selected_lfo + 1) % 6
            self.selected_lfo_start_value = self.get_clock_division()

    def get_clock_division(self):
        """Determine the new clock division based on the position of knob 2"""
        return k2.read_position(MAX_HARMONIC) + 1

    def reset(self):
        """Reset all LFOs to zero volts, maintaining their divisions"""
        self.degree = 0

    def get_delay_increment_value(self):
        """Calculate the wait time between degrees"""
        delay = (0.1 - (k1.read_position(100, 1) / 1000)) + (ain.read_voltage(1) / 100)
        return delay, round((((1 / delay) - 10) / 1) + 1)

    def save_state(self):
        """Save the current set of divisions to file"""
        # Prevent save-spam by limiting save rate to 1Hz
        if self.last_saved() < 1000:
             return

        self.state_needs_saving = False
        self.save_state_json({
            "divisions": self.divisions,
            "modes": self.modes,
            "viewAllWaveforms": self.viewAllWaveforms,
        })

    def update_display(self):
        """Update the OLED display every 10 cycles (degrees)"""
        oled.scroll(-1, 0)

        if round(self.degree, -1) % 10 == 0:
            oled.show()

    def increment(self):
        """Increment the current degree and determine new values of delay and increment_value"""
        self.degree += self.increment_value
        self.delay, self.increment_value = self.get_delay_increment_value()
        sleep_ms(int(self.delay))

    def draw_wave(self):
        shape = self.modes[self.selected_lfo]

        if shape == self.MODES_SHAPES['SINE']:
            oled.pixel(3, 31, 1)
            oled.pixel(3, 30, 1)
            oled.pixel(3, 29, 1)
            oled.pixel(4, 28, 1)
            oled.pixel(4, 27, 1)
            oled.pixel(4, 26, 1)
            oled.pixel(4, 25, 1)
            oled.pixel(5, 24, 1)
            oled.pixel(6, 23, 1)
            oled.pixel(7, 23, 1)
            oled.pixel(8, 24, 1)
            oled.pixel(9, 25, 1)
            oled.pixel(9, 26, 1)
            oled.pixel(9, 27, 1)
            oled.pixel(10, 28, 1)
            oled.pixel(10, 29, 1)
            oled.pixel(11, 30, 1)
            oled.pixel(12, 31, 1)
            oled.pixel(13, 31, 1)
            oled.pixel(14, 30, 1)
            oled.pixel(15, 29, 1)
            oled.pixel(15, 28, 1)
            oled.pixel(15, 27, 1)
            oled.pixel(15, 26, 1)
            oled.pixel(16, 25, 1)
            oled.pixel(16, 24, 1)
            oled.pixel(16, 23, 1)
        elif shape == self.MODES_SHAPES['SAW']:
            oled.line(3, 31, 9, 24, 1)
            oled.vline(9, 24, 8, 1)
            oled.line(9, 31, 15, 24, 1)
            oled.vline(15, 24, 8, 1)
        elif shape == self.MODES_SHAPES['SQUARE']:
            oled.vline(3, 24, 8, 1)
            oled.hline(3, 24, 6, 1)
            oled.vline(9, 24, 8, 1)
            oled.hline(9, 31, 6, 1)
            oled.vline(15, 24, 8, 1)
        elif shape == self.MODES_SHAPES['OFF']:
            oled.line(3, 24, 15, 31, 1)
            oled.line(15, 24, 3, 31, 1)
        elif shape == self.MODES_SHAPES['RANDOM']:
            oled.pixel(3, 29, 1)
            oled.pixel(4, 28, 1)
            oled.pixel(4, 27, 1)
            oled.pixel(5, 26, 1)
            oled.pixel(6, 26, 1)
            oled.pixel(7, 27, 1)
            oled.pixel(8, 28, 1)
            oled.pixel(9, 28, 1)
            oled.pixel(10, 27, 1)
            oled.pixel(10, 26, 1)
            oled.pixel(10, 25, 1)
            oled.pixel(11, 24, 1)
            oled.pixel(12, 25, 1)
            oled.pixel(13, 26, 1)
            oled.pixel(13, 27, 1)
            oled.pixel(14, 28, 1)
            oled.pixel(14, 29, 1)
            oled.pixel(15, 30, 1)
            oled.pixel(16, 30, 1)
        elif shape == self.MODES_SHAPES['NOISE']:
            oled.pixel(3, 26, 1)
            oled.pixel(3, 30, 1)
            oled.pixel(4, 24, 1)
            oled.pixel(4, 23, 1)
            oled.pixel(5, 31, 1)
            oled.pixel(5, 29, 1)
            oled.pixel(6, 25, 1)
            oled.pixel(6, 23, 1)
            oled.pixel(7, 29, 1)
            oled.pixel(7, 25, 1)
            oled.pixel(8, 26, 1)
            oled.pixel(8, 30, 1)
            oled.pixel(9, 27, 1)
            oled.pixel(9, 24, 1)
            oled.pixel(10, 26, 1)
            oled.pixel(10, 29, 1)
            oled.pixel(11, 24, 1)
            oled.pixel(11, 30, 1)
            oled.pixel(12, 23, 1)
            oled.pixel(12, 25, 1)
            oled.pixel(13, 29, 1)
            oled.pixel(13, 25, 1)
            oled.pixel(14, 28, 1)
            oled.pixel(14, 26, 1)
            oled.pixel(15, 24, 1)
            oled.pixel(15, 31, 1)
            oled.pixel(16, 27, 1)
            oled.pixel(16, 30, 1)

    def display_selected_lfo(self):
        """Draw the current LFO's number and division to the OLED display"""
        oled.fill_rect(0, 0, 20, 32, 0)
        oled.fill_rect(0, 0, 20, 9, 1)
        oled.text(f'{self.selected_lfo + 1}', 6, 1, 0)

        number = self.divisions[self.selected_lfo]
        x = 2 if number >= 10 else 6
        oled.text(f'{number}', x, 12, 1)

        self.draw_wave()

    def calculate_voltage(self, cv, multiplier):
        """Determine the voltage based on current degree, wave shape, and MAX_VOLTAGE"""
        three_sixty = 360 * multiplier
        degree_three_sixty = self.degree % three_sixty
        shape = self.modes[cvs.index(cv)]
        voltage = 0

        if shape == self.MODES_SHAPES['SINE']:
            voltage = (0 - (cos(self.rad * (1 / multiplier))) + 1) * (MAX_VOLTAGE / 2)
        elif shape == self.MODES_SHAPES['SAW']:
            voltage = (degree_three_sixty / three_sixty) * MAX_VOLTAGE
        elif shape == self.MODES_SHAPES['SQUARE']:
            if degree_three_sixty / three_sixty < 0.5:
                voltage = MAX_VOLTAGE
            else:
                voltage = 0
        elif shape == self.MODES_SHAPES['RANDOM']:  # This is NOT actually random, it is the sum of 3 out of sync sine waves, but it produces a fluctuating voltage that is near impossible to predict over time, and which can be clocked to be in time
            voltage = (((0 - (cos(self.rad * (1 / multiplier))) + 1) * (MAX_VOLTAGE / 2)) / 3) \
                    + (((0 - (cos(self.rad * (1 / (multiplier * 2.3)))) + 1) * (MAX_VOLTAGE / 2)) / 3) \
                    + (((0 - (cos(self.rad * (1 / (multiplier * 5.6)))) + 1) * (MAX_VOLTAGE / 2)) / 3)
        elif shape == self.MODES_SHAPES['NOISE']:  # The division knob is affecting the spread for the noise
            voltage = MAX_VOLTAGE * randint(0, int((1000 / MAX_HARMONIC) * multiplier)) / 1000

        return voltage

    def display_graphic_lines(self):
        """Draw the lines displaying each LFO's voltage to the OLED display"""
        self.rad = radians(self.degree)
        oled.vline(self.pixel_x, 0, OLED_HEIGHT, 0)

        for index, (cv, multiplier) in enumerate(zip(cvs, self.divisions)):
            volts = self.calculate_voltage(cv, multiplier)
            cv.voltage(volts)

            if self.viewAllWaveforms:
                oled.pixel(self.pixel_x, self.pixel_y - int(volts * (self.pixel_y / 10)), 1)
            elif index == self.selected_lfo:
                oled.pixel(self.pixel_x, self.pixel_y - int(volts * (self.pixel_y / 10)), 1)


    def check_change_clock_division(self):
        """Change current LFO's division with knob movement detection"""
        self.clock_division = self.get_clock_division()

        if self.clock_division != self.selected_lfo_start_value:
            self.selected_lfo_start_value = self.divisions[self.selected_lfo] = self.clock_division
            self.state_needs_saving = True

    def main(self):
        while True:
            self.check_change_clock_division()

            self.display_graphic_lines()

            self.display_selected_lfo()

            self.update_display()

            self.increment()

            if self.state_needs_saving:
                self.save_state()


if __name__ == "__main__":
    HarmonicLFOs().main()



# Copyright 2022 Allen Synthesis
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from europi import *
import machine
from rp2 import PIO, StateMachine, asm_pio
from europi_script import EuroPiScript

'''
Poly Square
author: Tyler Schreiber (github.com/t-schreibs)
date: 2022-05-10
labels: oscillator, poly

A poly square oscillator with detuning & several polyphony modes. The analog
input receives a V/oct and sets the root pitch of the 6 oscillators, and knob
1 sets the spread of the detuning across the oscillators. Knob 2 sets the
polyphony mode of the oscillators, which output on CVs 1-6.

digital_in: unused
analog_in: V/oct

knob_1: detune
knob_2: polyphony mode

button_1: while depressed, 'tuning mode' is turned on; this changes the knob functionality:
    knob_1: coarse tune (up to 8 octaves)
    knob_2: fine tune (up to a full octave swing)
button_2: toggles the maximum detune between a half step and a major 9th

output_1: oscillator 1
output_2: oscillator 2
output_3: oscillator 3
output_4: oscillator 4
output_5: oscillator 5
output_6: oscillator 6
'''

# Assembly code program for the PIO square oscillator
# Thanks to Ben Everard at HackSpace for the basis of this program:
# https://hackspace.raspberrypi.com/articles/raspberry-pi-picos-pio-for-mere-mortals-part-3-sound
@asm_pio(sideset_init=PIO.OUT_LOW)
def square_prog():
    # Initialize x & y variables - these are used to count down the
    # loops which set the length of the square wave's cycle
    label("restart")
    pull(noblock) .side(0)
    mov(x, osr)
    mov(y, isr)
    # Start loop
    # Here, the pin is low, and it will count down y
    # until y=x, then set the pin high and jump to the next section
    label("up_loop")
    jmp(x_not_y, "skip_up")
    nop()         .side(1)
    jmp("down")
    label("skip_up")
    jmp(y_dec, "up_loop")
    # Mirror the above loop, but with the pin high to form the second
    # half of the square wave
    label("down")
    mov(y, isr)
    label("down_loop")
    jmp(x_not_y, "skip_down")
    nop() .side(0)
    jmp("restart")
    label("skip_down")
    jmp(y_dec, "down_loop")

# Class for managing the settings for a polyphony mode
class PolyphonyMode:
    def __init__(self, name, voltage_offsets):
        self.name = name
        self.voltage_offsets = voltage_offsets

# Class for managing a state machine running the PIO oscillator program
class SquareOscillator:
    def __init__(self, sm_id, pin, max_count, count_freq):
        self._sm = StateMachine(
            sm_id, square_prog, freq=2 * count_freq, sideset_base=Pin(pin))
        # Use exec() to load max count into ISR
        self._sm.put(max_count)
        self._sm.exec("pull()")
        self._sm.exec("mov(isr, osr)")
        self._sm.active(1)
        self._max_count = max_count
        self._count_freq = count_freq

    def set(self, value):
        # Minimum value is -1 (completely turn off), 0 actually still
        # produces a narrow pulse
        value = clamp(value, -1, self._max_count)
        self._sm.put(value)

    # Converts Hertz to the value the state machine running the PIO
    # program needs
    def get_pitch(self, hertz):
        return int( -1 * (((self._count_freq / hertz) -
            (self._max_count * 4))/4))

class KnobState:
    def __init__(self, k1, k2):
        self.k1 = k1
        self.k2 = k2

class PolySquare(EuroPiScript):
    def __init__(self):
        k1.set_samples(256)
        k2.set_samples(256)
        # PIO settings
        max_count = 1_000_000
        count_freq = 50_000_000
        # Thanks to djmjr (github.com/djmjr) for testing & determining that
        # 6 oscillators can be run simultanously without any issues:
        # https://github.com/djmjr/EuroPi/blob/poly-squares-mods/software/contrib/poly_square_mods.py
        self.oscillators = [
            SquareOscillator(0, 21, max_count, count_freq),
            SquareOscillator(1, 20, max_count, count_freq),
            SquareOscillator(2, 16, max_count, count_freq),
            SquareOscillator(3, 17, max_count, count_freq),
            SquareOscillator(4, 18, max_count, count_freq),
            SquareOscillator(5, 19, max_count, count_freq)
        ]
        # To add more polyphony modes, include them in this list. The offsets
        # are V/oct offsets (ie, a 5th = 7/12, an octave = 1, etc.). If the number
        # of offsets in the tuple doesn't match the length of the self.oscillators
        # list above, oscillators will wrap back to the first offset (ie, if there
        # are 3 offsets, and 6 oscillators, the fourth oscillator will take the
        # first offset, the fifth will take the second, etc.).
        self.modes = [
            PolyphonyMode("Unison", (0,)),
            PolyphonyMode("5th", (0, 0, 7/12)),
            PolyphonyMode("Octave", (0, 0, 1)),
            PolyphonyMode("Power chord", (0, 7/12, 1)),
            PolyphonyMode("Stacked 5ths", (0, 7/12, 14/12)),
            PolyphonyMode("Minor triad", (0, 7/12, 15/12)),
            PolyphonyMode("Major triad", (0, 7/12, 16/12)),
            PolyphonyMode("Diminished", (0, 6/12, 15/12)),
            PolyphonyMode("Augmented", (0, 8/12, 16/12)),
            PolyphonyMode("Major 6th", (0, 4/12, 9/12)),
            PolyphonyMode("Major 7th", (0, 4/12, 11/12)),
            PolyphonyMode("Minor 7th", (0, 3/12, 10/12)),
            PolyphonyMode("Major penta.", (0, 2/12, 4/12, 7/12, 9/12, 1)),
            PolyphonyMode("Minor penta.", (0, 2/12, 3/12, 7/12, 9/12, 1)),
            PolyphonyMode("Whole tone", (0, 2/12, 4/12, 6/12, 8/12, 10/12))
        ]
        self.current_mode = None
        self.ui_update_requested = True
        self.save_state_requested = False
        self.detune_amount = None
        self.coarse_tune = 0
        self.fine_tune = .5
        self.tuning_mode = False
        self.max_detune = 1
        self.tuning_mode_compare_knob_state = KnobState(
            None,
            None
        )
        self.load_state()

        @b1.handler
        def tuning_mode_on():
            self.tuning_mode_compare_knob_state = KnobState(
                k1.percent(),
                k2.percent()
            )
            self.tuning_mode = True
            self.ui_update_requested = True

        @b1.handler_falling
        def tuning_mode_off():
            self.tuning_mode = False
            self.ui_update_requested = True
            # Save the tuning settings after the button is released
            self.save_state_requested = True

        # self.max_detune is not a boolean value, in case the values need to be
        # updated in future or more than two values become available in the UI
        # (by, say, mapping it to a knob instead of a button toggle)
        @b2.handler
        def change_max_detune():
            if self.max_detune == 1:
                self.max_detune = 14
            else:
                self.max_detune = 1
            self.ui_update_requested = True
            self.save_state_requested = True

    # Converts V/oct signal to Hertz, with 0V = C0
    def get_hertz(self, voltage):
        # Start with A0 because it's a nice, rational number
        a0 = 27.5
        # Subtract 3/4 from the voltage value so that 0V = C0
        return a0 * 2 ** (voltage - 3/4)

    # Returns the linear step distance between elements such that all are
    # equally spaced. Ex, for 6 elements between 0 and 100 (inclusive), the
    # step would be 20 (the elements would then be 0, 20, 40, 60, 80, 100).
    def get_step_distance(self, first, last, count):
        return (last - first) / (count - 1)

    # Returns the total voltage offset of the current tuning values
    def get_tuning(self):
        return self.coarse_tune * 8 + self.fine_tune - .5

    # Through-zero detuning of current oscillator based on its position in
    # the list of oscillators - the central oscillator(s) stay(s) in tune while
    # the outer oscillators are progressively detuned.
    def get_detuning(self, detune_amount, oscillator_index):
        oscillator_count = len(self.oscillators)
        step_distance = self.get_step_distance(
            0, self.max_detune, oscillator_count)
        return detune_amount * step_distance * (oscillator_index -
             (oscillator_count - 1) / 2)

    # Returns a voltage offset for the oscillator based on the current
    # polyphony mode. This allows for things like triads, with the offsets
    # set accordingly.
    def get_offset(self, oscillator_index):
        mode = self.modes[self.current_mode]
        return mode.voltage_offsets[oscillator_index %
            len(mode.voltage_offsets)]

    # Saves oscillator tuning & detuning settings
    def save_state(self):
        settings = {
            "c": self.coarse_tune,
            "f": self.fine_tune,
            "m": self.max_detune
        }
        self.save_state_json(settings)
        self.save_state_requested = False

    # Loads oscillator tuning & detuning settings
    def load_state(self):
        settings = self.load_state_json()
        if "c" in settings:
            self.coarse_tune = settings["c"]
        if "f" in settings:
            self.fine_tune = settings["f"]
        if "m" in settings:
            self.max_detune = settings["m"]

    # Draws the UI for the "tuning" mode
    def draw_tuning_ui(self):
        oled.fill(0)
        padding = 2
        line_height = 9
        title = "Tuning"
        # Center the title at the top of the screen
        title_x = int((OLED_WIDTH - ((len(title) + 1) * 7)) / 2) - 1
        oled.text(title, title_x, padding)
        tuning_bar_x = 60
        tuning_bar_width = OLED_WIDTH - tuning_bar_x - padding
        # Coarse tuning bar
        oled.text("coarse:", padding, padding + line_height)
        oled.rect(tuning_bar_x, padding + line_height, tuning_bar_width, 8, 1)
        oled.fill_rect(
            tuning_bar_x,
            padding + line_height,
            int(self.coarse_tune * tuning_bar_width), 8, 1)
        # Fine tuning bar
        oled.text("fine:", padding + 16, padding + line_height * 2)
        oled.rect(
            tuning_bar_x, padding + line_height * 2, tuning_bar_width, 8, 1)
        if self.fine_tune < 0.5:
            filled_bar_width = int((0.5 - self.fine_tune) * tuning_bar_width)
            oled.fill_rect(
                int(tuning_bar_x + tuning_bar_width / 2 - filled_bar_width),
                padding + line_height * 2, filled_bar_width, 8, 1)
        elif self.fine_tune == 0.5:
            line_x = int(tuning_bar_x + tuning_bar_width / 2)
            oled.vline(line_x, padding + line_height * 2, 8, 1)
        else:
            oled.fill_rect(
                int(tuning_bar_x + tuning_bar_width / 2 + 2),
                padding + line_height * 2,
                int((self.fine_tune - 0.5) * tuning_bar_width), 8, 1)
        oled.show()

    # Draws the default UI
    def draw_main_ui(self):
        oled.centre_text(self.modes[self.current_mode].name +
            "\nmax. detune: " + str(self.max_detune))

    def numbers_are_close(self, current, other, allowed_error):
        if current == None or other == None:
            return False
        return abs(current - other) <= allowed_error

    def update_ui(self):
        if self.tuning_mode:
            self.draw_tuning_ui()
        else:
            self.draw_main_ui()
        self.ui_update_requested = False

    def update_tuning_settings(self):
        new_coarse_tune = k1.percent()
        new_fine_tune = k2.percent()
        allowed_error = 0.005
        # Only update the coarse or fine tuning setting if the knob has
        # been moved since button 1 was depressed - thanks to djmjr
        # (github.com/djmjr) for the idea:
        # https://github.com/djmjr/EuroPi/blob/poly-squares-mods/software/contrib/poly_square_mods.py
        if not (self.numbers_are_close(
                new_coarse_tune, self.tuning_mode_compare_knob_state.k1,
                allowed_error) or new_coarse_tune == self.coarse_tune):
            self.coarse_tune = new_coarse_tune
            self.tuning_mode_compare_knob_state.k1 = None
            self.ui_update_requested = True
        if not (self.numbers_are_close(
                new_fine_tune, self.tuning_mode_compare_knob_state.k2,
                allowed_error) or new_fine_tune == self.fine_tune):
            self.fine_tune = new_fine_tune
            self.tuning_mode_compare_knob_state.k2 = None
            self.ui_update_requested = True

    # Updates oscillator settings based on the current knob positions &
    # analog input
    def update_settings(self):
        analog_input = ain.read_voltage(32)
        if self.tuning_mode:
            self.update_tuning_settings()
        else:
            self.detune_amount = k1.percent() / 12
            new_mode = k2.read_position(len(self.modes))
            if not new_mode == self.current_mode:
                self.ui_update_requested = True
                self.current_mode = new_mode
        for oscillator in self.oscillators:
            # Add up the V/oct from the analog input, the offset from the
            # polyphony mode, the adjustment from the tuning, and the
            # adjustment from the detune amount to get the final pitch for
            # the oscillator
            oscillator_index = self.oscillators.index(oscillator)
            oscillator.set(oscillator.get_pitch(
                self.get_hertz(
                    analog_input + self.get_offset(oscillator_index) +
                    self.get_tuning() +
                    self.get_detuning(self.detune_amount, oscillator_index))))

    def main(self):
        while True:
            self.update_settings()
            if self.ui_update_requested:
                self.update_ui()
            if self.save_state_requested:
                self.save_state()

# Main script execution
if __name__ == '__main__':
    PolySquare().main()
